from mosaic_utils.ai.encoding_utils import find_categorical
"""Weight of Evidence"""
import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator
from category_encoders.ordinal import OrdinalEncoder
import category_encoders.utils as util
from sklearn.utils.random import check_random_state

class WOEEncoder(BaseEstimator, util.TransformerWithTargetMixin):
    """Weight of Evidence coding for categorical features.

    Parameters
    ----------

    verbose: int
        integer indicating verbosity of the output. 0 for none.
    cols: list
        a list of columns to encode, if None, all string columns will be encoded.
    drop_invariant: bool
        boolean for whether or not to drop columns with 0 variance.
    return_df: bool
        boolean for whether to return a pandas DataFrame from transform (otherwise it will be a numpy array).
    handle_missing: str
        options are 'return_nan', 'error' and 'value', defaults to 'value', which will assume WOE=0.
    handle_unknown: str
        options are 'return_nan', 'error' and 'value', defaults to 'value', which will assume WOE=0.
    randomized: bool,
        adds normal (Gaussian) distribution noise into training data in order to decrease overfitting (testing data are untouched).
    sigma: float
        standard deviation (spread or "width") of the normal distribution.
    regularization: float
        the purpose of regularization is mostly to prevent division by zero.
        When regularization is 0, you may encounter division by zero.

    Example
    -------
    >>> from category_encoders import *
    >>> import pandas as pd
    >>> from sklearn.datasets import load_boston
    >>> bunch = load_boston()
    >>> y = bunch.target > 22.5
    >>> X = pd.DataFrame(bunch.data, columns=bunch.feature_names)
    >>> enc = WOEEncoder(cols=['CHAS', 'RAD']).fit(X, y)
    >>> numeric_dataset = enc.transform(X)
    >>> print(numeric_dataset.info())
    <class 'pandas.core.frame.DataFrame'>
    RangeIndex: 506 entries, 0 to 505
    Data columns (total 13 columns):
    CRIM       506 non-null float64
    ZN         506 non-null float64
    INDUS      506 non-null float64
    CHAS       506 non-null float64
    NOX        506 non-null float64
    RM         506 non-null float64
    AGE        506 non-null float64
    DIS        506 non-null float64
    RAD        506 non-null float64
    TAX        506 non-null float64
    PTRATIO    506 non-null float64
    B          506 non-null float64
    LSTAT      506 non-null float64
    dtypes: float64(13)
    memory usage: 51.5 KB
    None

    References
    ----------

    .. [1] Weight of Evidence (WOE) and Information Value Explained, from
    https://www.listendata.com/2015/03/weight-of-evidence-woe-and-information.html

    """

    def __init__(self, verbose=0, cols=None, drop_invariant=False, return_df=True,
                 handle_unknown='value', handle_missing='value', random_state=None, randomized=False, sigma=0.05, regularization=1.0):
        self.verbose = verbose
        self.return_df = return_df
        self.drop_invariant = drop_invariant
        self.drop_cols = []
        self.cols = cols
        self.ordinal_encoder = None
        self._dim = None
        self.mapping = None
        self.handle_unknown = handle_unknown
        self.handle_missing = handle_missing
        self._sum = None
        self._count = None
        self.random_state = random_state
        self.randomized = randomized
        self.sigma = sigma
        self.regularization = regularization
        self.feature_names = None

    def fit(self, X, y, **kwargs):
        """Fit encoder according to X and binary y.

        Parameters
        ----------

        X : array-like, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number of samples
            and n_features is the number of features.
        y : array-like, shape = [n_samples]
            Binary target values.

        Returns
        -------

        self : encoder
            Returns self.

        """

        # Unite parameters into pandas types
        X = util.convert_input(X)
        y = util.convert_input_vector(y, X.index).astype(float)

        # The lengths must be equal
        if X.shape[0] != y.shape[0]:
            raise ValueError("The length of X is " + str(X.shape[0]) + " but length of y is " + str(y.shape[0]) + ".")

        # The label must be binary with values {0,1}
        unique = y.unique()
        if len(unique) != 2:
            raise ValueError("The target column y must be binary. But the target contains " + str(len(unique)) + " unique value(s).")
        if y.isnull().any():
            raise ValueError("The target column y must not contain missing values.")
        if np.max(unique) < 1:
            raise ValueError("The target column y must be binary with values {0, 1}. Value 1 was not found in the target.")
        if np.min(unique) > 0:
            raise ValueError("The target column y must be binary with values {0, 1}. Value 0 was not found in the target.")

        self._dim = X.shape[1]
        self.cols,_ = find_categorical(X,10)
        # If columns aren't passed, just use every string column
        if self.cols is None:
            self.cols = util.get_obj_cols(X)
        else:
            self.cols = util.convert_cols_to_list(self.cols)

        if self.handle_missing == 'error':
            if X[self.cols].isnull().any().any():
                raise ValueError('Columns to be encoded can not contain null')

        self.ordinal_encoder = OrdinalEncoder(
            verbose=self.verbose,
            cols=self.cols,
            handle_unknown='value',
            handle_missing='value'
        )
        self.ordinal_encoder = self.ordinal_encoder.fit(X)
        X_ordinal = self.ordinal_encoder.transform(X)

        # Training
        self.mapping = self._train(X_ordinal, y)

        X_temp = self.transform(X, override_return_df=True)
        self.feature_names = X_temp.columns.tolist()

        # Store column names with approximately constant variance on the training data
        if self.drop_invariant:
            self.drop_cols = []
            generated_cols = util.get_generated_cols(X, X_temp, self.cols)
            self.drop_cols = [x for x in generated_cols if X_temp[x].var() <= 10e-5]
            try:
                [self.feature_names.remove(x) for x in self.drop_cols]
            except KeyError as e:
                if self.verbose > 0:
                    print("Could not remove column from feature names."
                    "Not found in generated cols.\n{}".format(e))
        return self

    def transform(self, X, y=None, override_return_df=False):
        """Perform the transformation to new categorical data. When the data are used for model training,
        it is important to also pass the target in order to apply leave one out.

        Parameters
        ----------

        X : array-like, shape = [n_samples, n_features]
        y : array-like, shape = [n_samples] when transform by leave one out
            None, when transform without target information (such as transform test set)

        Returns
        -------

        p : array, shape = [n_samples, n_numeric + N]
            Transformed values with encoding applied.

        """

        if self.handle_missing == 'error':
            if X[self.cols].isnull().any().any():
                raise ValueError('Columns to be encoded can not contain null')

        if self._dim is None:
            raise ValueError('Must train encoder before it can be used to transform data.')

        # Unite the input into pandas types
        X = util.convert_input(X)

        # Then make sure that it is the right size
        if X.shape[1] != self._dim:
            raise ValueError('Unexpected input dimension %d, expected %d' % (X.shape[1], self._dim,))

        # If we are encoding the training data, we have to check the target
        if y is not None:
            y = util.convert_input_vector(y, X.index).astype(float)
            if X.shape[0] != y.shape[0]:
                raise ValueError("The length of X is " + str(X.shape[0]) + " but length of y is " + str(y.shape[0]) + ".")

        if not list(self.cols):
            return X

        # Do not modify the input argument
        X = X.copy(deep=True)

        X = self.ordinal_encoder.transform(X)

        if self.handle_unknown == 'error':
            if X[self.cols].isin([-1]).any().any():
                raise ValueError('Unexpected categories found in dataframe')

        # Loop over columns and replace nominal values with WOE
        X = self._score(X, y)

        # Postprocessing
        # Note: We should not even convert these columns.
        if self.drop_invariant:
            for col in self.drop_cols:
                X.drop(col, 1, inplace=True)

        if self.return_df or override_return_df:
            return X
        else:
            return X.values

    def _train(self, X, y):
        # Initialize the output
        mapping = {}

        # Calculate global statistics
        self._sum = y.sum()
        self._count = y.count()

        for switch in self.ordinal_encoder.category_mapping:
            col = switch.get('col')
            values = switch.get('mapping')
            # Calculate sum and count of the target for each unique value in the feature col
            stats = y.groupby(X[col]).agg(['sum', 'count'])  # Count of x_{i,+} and x_i

            # Create a new column with regularized WOE.
            # Regularization helps to avoid division by zero.
            # Pre-calculate WOEs because logarithms are slow.
            nominator = (stats['sum'] + self.regularization) / (self._sum + 2*self.regularization)
            denominator = ((stats['count'] - stats['sum']) + self.regularization) / (self._count - self._sum + 2*self.regularization)
            woe = np.log(nominator / denominator)

            # Ignore unique values. This helps to prevent overfitting on id-like columns.
            woe[stats['count'] == 1] = 0

            if self.handle_unknown == 'return_nan':
                woe.loc[-1] = np.nan
            elif self.handle_unknown == 'value':
                woe.loc[-1] = 0

            if self.handle_missing == 'return_nan':
                woe.loc[values.loc[np.nan]] = np.nan
            elif self.handle_missing == 'value':
                woe.loc[-2] = 0

            # Store WOE for transform() function
            mapping[col] = woe

        return mapping

    def _score(self, X, y):
        for col in self.cols:
            # Score the column
            X[col] = X[col].map(self.mapping[col])

            # Randomization is meaningful only for training data -> we do it only if y is present
            if self.randomized and y is not None:
                random_state_generator = check_random_state(self.random_state)
                X[col] = (X[col] * random_state_generator.normal(1., self.sigma, X[col].shape[0]))

        return X

    def get_feature_names(self):
        """
        Returns the names of all transformed / added columns.

        Returns
        -------
        feature_names: list
            A list with all feature names transformed or added.
            Note: potentially dropped features are not included!

        """
        if not isinstance(self.feature_names, list):
            raise ValueError("Estimator has to be fitted to return feature names.")
        else:
            return self.feature_names



# -*- coding: utf-8 -*-
import requests
from flask import jsonify
from .contants import InputParamPath, InputParamReferenceType


def get_all_input_params(env, headers, url, reference_id, reference_type):
    """
    Get all level input params
    reference type: MODEL, PROJECT, USER, NOTEBOOK
    """
    try:
        param_url = url + InputParamPath.input_params_path.format(
            reference_id=reference_id, inputReferenceType=reference_type
        )
        response = requests.get(param_url, headers=headers)

        if reference_type == InputParamReferenceType.ref_type_model:
            if response.status_code == 200:
                list_of_params = response.json()
                if list_of_params:
                    for param in list_of_params:
                        if param["inputParameterType"] != "SYSTEM":
                            env[str(param["inputParameterName"])] = str(
                                param["parameterValue"]
                            )
            return env
        else:
            return create_param_dict(env, response)

    except Exception:
        return []


@DeprecationWarning
def get_project_params(env, headers, url, project_id):
    """
    Get project level input params
    """
    try:
        param_url = url + InputParamPath.project.format(project_id=project_id)
        response = requests.get(param_url, headers=headers)
        return create_param_dict(env, response)
    except Exception:
        raise


@DeprecationWarning
def get_notebook_params(env, headers, url, notebook_id):
    """
    Get notebook level input params
    """
    try:
        param_url = url + InputParamPath.notebook.format(notebook_id=notebook_id)
        response = requests.get(param_url, headers=headers)
        return create_param_dict(env, response)
    except Exception:
        raise


@DeprecationWarning
def get_user_params(env, headers, url):
    """
    Get user level input params
    """
    try:
        param_url = url + InputParamPath.user
        response = requests.get(param_url, headers=headers)
        return create_param_dict(env, response)
    except Exception:
        raise


@DeprecationWarning
def get_model_params(env, headers, url, version_id):
    """
    Get model level input params
    """
    try:
        param_url = url + InputParamPath.model.format(version_id=version_id)
        response = requests.get(param_url, headers=headers)
        if response.status_code == 200:
            list_of_params = response.json()
            if list_of_params:
                for param in list_of_params:
                    if param["inputParameterType"] != "SYSTEM":
                        env[str(param["inputParameterName"])] = str(
                            param["parameterValue"]
                        )
        return env
    except Exception:
        raise


def create_param_dict(env, response):
    if response.status_code == 200:
        list_of_params = response.json()
        if list_of_params:
            for param in list_of_params:
                key_val = {
                    remove_special_char(param["inputParameterName"]): param[
                        "parameterValue"
                    ]
                }
                env.update(key_val)
    return env


def remove_special_char(var):
    """
    Method to remove spaces & special characters apart from '-' & "_" from var,
    returns only alphanumeric values
    :param var: variable name
    """
    return "".join(e for e in var if e == "-" or e == "_" or e.isalnum())



class Headers:
    authorization = "Authorization"
    x_enabled_repo = "Enabled-Repo"
    x_auth_userid = "X-Auth-Userid"
    x_auth_username = "X-Auth-Username"
    x_auth_email = "X-Auth-Email"
    x_project_id = "X-Project-Id"
    x_request_id = "X-Request-Id"
    x_repo_id = "X-Repo-Id"
    x_branch_name = "X-Branch-Name"
    x_default_repo_flag = "X_Default_Repo_Flag"
    x_product_id = "X-Product-Id"
    x_tenant_id = "X-Auth-TenantId"
    x_auth_realm_name = "X-Auth-RealmName"
    access_control_allow_origin = "Access-Control-Allow-Origin"
    access_control_allow_methods = "Access-Control-Allow-Methods"
    access_control_allow_headers = "Access-Control-Allow-Headers"
    x_cli_script = "X-Cli-Script"



from mosaic_utils.ai.score.base import ScoreBase
from typing import Tuple, Union, List, Any
import tensorflow as tf
import numpy as np


class ScoreTemplateExample(ScoreBase):
    """
    This Class Demonstrate How To Implements ScoreBase Interface Class And It Basic Usage.
    """
    def request_processing_fn(self, request) -> Tuple[int,
                                                      Union[np.ndarray,
                                                            List[np.ndarray],
                                                            tf.Tensor,
                                                            List[tf.Tensor],
                                                      ]]:
        """
        Processes Request Object -> Input data. It could be:
                A Numpy array (or array-like), or a list of arrays (in case the model has multiple inputs).
                A TensorFlow tensor, or a list of tensors (in case the model has multiple inputs).
                A dict mapping input names to the corresponding array/tensors, if the model has named inputs.

        :return: (n_inputs, payload's)
        """
        final_payload = []
        raw_payload = request.json

        for each_payload in raw_payload:
            final_payload.append(np.asarray(each_payload))
        print("Processed Payload", (len(final_payload), final_payload))

        return (len(final_payload), final_payload)

    def pre_processing_fn(self,
                          payload: Union[np.ndarray, tf.Tensor]) -> Union[np.ndarray, tf.Tensor]:
        """
        Accepts Single Payload At a time and applies pre_processing_step before prediction
        :param each_payload: Single Payload
        :return: Preprocessed Single Payload
        """
        # Not Doing Any Preprocessing Hence Returned payload
        return payload

    def prediction_fn(self,
                      model: Any,
                      pre_processed_input: Union[np.ndarray, tf.Tensor]
                      ):
        """
                Does the main prediction on pre_processed_input using supplied model.

                :param model: Supported Model
                :param pre_processed_input: Single Preprocessed Payload
                :return: Prediction Value From the model
        """

        return model.predict(pre_processed_input)

    class Meta:
        # List of Callables() can be attached For Calling After And Before Scoring
        pre_call_hooks = []
        post_call_hooks = []


if __name__ == "__main__":
    class FlaskRequest:
        """This is A Dummy Request Mocking Input"""
        json = [[12, 5, 2, 4],
                [7, 6, 8, 8],
                [1, 6, 7, 7]]

    from random import randint

    class MockModel:
        """This is A Dummy Model Mocking Model Prediction"""
        def predict(self, input):
            return randint(0, 9)

    obj = ScoreTemplateExample()
    model_predictions = obj.score(MockModel(), FlaskRequest(), dry_run=True)

    print("Model Prediction")
    print(model_predictions)

    print("Model Prediction Encoding")
    import json
    from mosaic_utils.ai.score.encoder import ScoreResponseDecoder, ScoreResponseEncoder
    d = json.dumps(model_predictions, cls=ScoreResponseEncoder)
    print(d)

    print("Model Prediction Decoding")
    e = json.loads(d, cls=ScoreResponseDecoder)
    print(e)




# -*- coding: utf-8 -*-
from rpy2.robjects import r
import rpy2.robjects as robjects
import rpy2.robjects.packages as rpackages
import json


def load_model(model_path):
    readRDS = robjects.r["readRDS"]
    return readRDS(model_path)


def dump_model(model, path):
    r.saveRDS(model, path)


def load_r_packages(package_list):
    if package_list:
        package_list = f"('{package_list[0]}')" if len(package_list) == 1 else tuple(package_list)
        load_func = f"""function(){{
                        packages <- c{package_list}
                        for (package in packages){{
                            library(package, character.only = TRUE)
                        }}}}"""
        try:
            robjects.r(load_func)()
            print('Successfully loaded packages')
        except Exception as ex:
            print('Error while loading packages: ', ex)


def get_r_packages():
    fetch_package_name = """
        function(){
            packages <- installed.packages()
            df <- data.frame(packages)
            package_name <- sapply(df$Package, as.character)
            return(package_name)
        }
    """
    fetch_version_name = """
        function(){
            packages <- installed.packages()
            df <- data.frame(packages)
            package_version <- sapply(df$Version, as.character)
            return(package_version)
        }
    """
    package_name_r = robjects.r(fetch_package_name)
    version_name_r = robjects.r(fetch_version_name)

    packages = package_name_r()
    versions = version_name_r()
    package_list = list(packages)
    version_list = list(versions)
    return package_list, version_list


def load_r_json(val):
    rjson = rpackages.importr("rjson")
    r_json_val = rjson.fromJSON(json.dumps(val))
    return r_json_val


def check_pre_installed_packages(packages_list, version_list, cran_package_list):
    packages = []
    for i, package in enumerate(packages_list):
        if package not in cran_package_list:
            packages.append({"name": package, "version": version_list[i]})

    if len(packages) == 0:
        packages.append({"name": "devtools", "version": "2.3.2"})
        packages.append({"name": "versions", "version": "0.3"})

    return packages



# -*- coding: utf-8 -*-
import os


class MosaicAI:
    server = os.getenv("MOSAIC_AI_SERVER", "http://localhost:5000/registry/api")


class MLModelMetricsV1:
    lc = "/v1/ml-model/metrics"
    rud = "/v1/ml-model/metrics/{version_id}/{tag}"
    u = "/v1/pipeline/{pipeline_id}/tag/{tag}"


class Client:
    config_file = "~/.mosaic.ai"


class Headers:
    authorization = "Authorization"
    x_project_id = "X-Project-Id"


class MLModelType:
    classification = "classification"
    regression = "regression"



# -*- coding: utf-8 -*-
from json import JSONEncoder
import numpy as np


class NumpyArrayEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, set):
            return list(obj)
        if isinstance(obj, np.ndarray) and obj.size >= 10:
            return None
        if isinstance(obj, np.ndarray) and obj.size <= 10:
            return obj.tolist()
        else:
            return str(obj)



import subprocess
import pandas as pd
import os
import random


def run_os_level_command(command, switch_sudo_user=False):
    """ This function calls os level command"""
    shell_type = ["sudo", "bash", "-c", command] if switch_sudo_user else ["sh", "-c", command]
    output = subprocess.run(shell_type, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                            universal_newlines=True)
    if output.returncode != 0:
        print(output)
        return False
    return True


def verify_credentials(username, password):
    """Verify credentials"""
    command = "echo '{1}' | kinit {0}".format(username, password)
    is_valid_credentials = run_os_level_command(command)
    return is_valid_credentials


# def copy_kdm_artifact(kdm_directory, kdm_filename):
#     """Generate ticket for kerberos"""
#     command = "sudo cp {0}/{1} /etc/".format(kdm_directory, kdm_filename)
#     is_valid_ticket = run_os_level_command(command)
#     return is_valid_ticket


def create_new_host(kerberos_service_ip, host_name):
    """Copy KERBEROS_SERVICE_IP in hosts file"""
    command = "sudo su; sudo echo '{0} {1}' >> /etc/hosts;".format(kerberos_service_ip, host_name)
    switch_sudo_user = True
    is_file_copied = run_os_level_command(command, switch_sudo_user)
    return is_file_copied


def read_hive_configuration(kdm_directory, hive_config_file):
    """ read hive connection details"""
    df = pd.read_json(r"{0}/{1}".format(kdm_directory, hive_config_file), typ='series')
    return df.to_dict()


def set_environment_value(var_name, var_value):
    """export variable"""
    os.environ[var_name] = var_value


def get_ads_data(sql_query, database_name, username, password,
                 kdm_directory='/etc', hive_config_file='hive_config.json', row_size=None, source='hive',
                 compress=False):
    """
    This function generates auth ticket and returns result in dataframe
    arguments it accepts
    sql_query : operation to be performed specified by sql_query
    database_name: database to connect
    username: username of kerberos
    password: password of kerberos
    kdm_directory : default directory location of kdc file is /data,
    you can specify other location
    kdm_filename: default kdc file name is krb5.conf,
    you can specify other filename
    row_size : number of records to fetch, default value is 500000
    """

    # Generate principle
    is_valid_credentials = verify_credentials(username, password)
    hive_config = read_hive_configuration(kdm_directory, hive_config_file)
    #is_host_appended = create_new_host(hive_config.get("KERBEROS_SERVICE_IP"),
    #                            hive_config.get("KEREROS_SERVICE_ALIAS"))

    if is_valid_credentials and hive_config:
        conn = get_hive_connection(hive_config, source, database_name)
        conn.execute(sql_query)
        df = pd.DataFrame()
        row_size = 500000 if row_size is None else row_size
        data = conn.fetchmany(row_size)
        field_names = [i[0] for i in conn.description]
        while data:
            df1 = pd.DataFrame(data, columns=field_names)
            if compress:
                df1 = squeeze_dataframe(df1)
            df = pd.concat([df, df1])
            del df1
            data = conn.fetchmany(row_size)
        return df


def get_hive_connection(hive_config, source, database_name):
    """
    Get hive connection cursor based on source type
    :param hive_config: dict having hive configuration
    :param source: create hive connection directly to hive host or get hive connection details from zookeeper
    :param database_name: hive database name
    :return: hive connection cursor
    """
    from pyhive import hive
    if source == "hive":
        # set KEREROS_SERVICE_ALIAS in environment variable, will be used by pyhive library
        set_environment_value('HIVE_HOSTNAME', hive_config.get("KEREROS_SERVICE_ALIAS"))
        return hive.connect(host=hive_config.get("KEREROS_SERVICE_ALIAS"),
                            port=int(hive_config.get("KERBEROS_PORT")),
                            auth=hive_config.get("KERBEROS_AUTH"),
                            kerberos_service_name=hive_config.get("KERBEROS_SERVICE_NAME"),
                            username=hive_config.get("KERBEROS_HIVE_USERNAME"),
                            scheme=hive_config.get("KERBEROS_HTTP_SCHEME"),
                            database=database_name).cursor()
    elif source == "zookeeper":
        zookeeper_host = hive_config.get("ZOOKEEPER_HOST")
        zookeeper_namespace = hive_config.get("ZOOKEEPER_NAMESPACE")
        host_list = get_host_list(zookeeper_host, zookeeper_namespace)
        host_length = host_list.__len__()
        random.seed()
        is_connected = False
        while is_connected is False and host_length > 0:
            index = random.randint(0, host_length - 1)
            host_details = host_list.pop(index)
            host = host_details.split(":")[0]
            port = host_details.split(":")[1]
            try:
                # set KEREROS_SERVICE_ALIAS in environment variable, will be used by pyhive library
                set_environment_value('HIVE_HOSTNAME', host)
                cursor = hive.connect(host=host,
                                      port=port,
                                      auth=hive_config.get("KERBEROS_AUTH"),
                                      kerberos_service_name=hive_config.get("KERBEROS_SERVICE_NAME"),
                                      username=hive_config.get("KERBEROS_HIVE_USERNAME"),
                                      scheme=hive_config.get("KERBEROS_HTTP_SCHEME"),
                                      database=database_name
                                      ).cursor()
                is_connected = True
            except Exception:
                is_connected = False
                if host_length > 1:
                    print(
                        "Can not connect " + host + ":" + port + " .try another server...")
                else:
                    print(
                        "Can not connect " + host + ":" + port + ", please check the connection config and the hiveserver")
                    raise Exception
            host_length -= 1
        return cursor


def get_host_list(host, namespace):
    """
    Get host list from zookeeper
    :param host: zookeeper URI
    :param namespace: zookeeper namespace
    :return: list of connection strings from that namespace
    """
    from kazoo.client import KazooClient
    from kazoo.handlers.threading import KazooTimeoutError
    try:
        client = KazooClient(hosts=host)
        client.start()
        node_children = client.get_children(namespace)
        host_list = []
        for hiveserver in node_children:
            server_string = hiveserver.split(';')[0].split('=')[1]
            host_list.append(server_string)
        client.stop()
        if len(host_list) == 0:
            raise Exception("No hive instance present in this zookeeper namespace")
        return host_list
    except KazooTimeoutError:
        print("Not able to connect to zookeeper server")
        raise
    except Exception as ex:
        print(ex)
        raise


def squeeze_dataframe(df):
    """
    Squeeze dataframe by reducing datatype to minimum
    :param df: dataframe to be squeeze
    :type df: pandas dataframe
    :return: squeezed dataframe
    :rtype: pandas dataframe
    """
    from pandas.api.types import CategoricalDtype
    cols = dict(df.dtypes)
    # ----- Check each column's type downcast or categorize as appropriate
    for col, type in cols.items():
        if type == 'float64':
            df[col] = pd.to_numeric(df[col], downcast='float')
        elif type == 'int64':
            df[col] = pd.to_numeric(df[col], downcast='integer')
        elif type == 'object':
            df[col] = df[col].astype(CategoricalDtype(ordered=True))
    return df



# -*- coding: utf-8 -*-
import os
import io
import tempfile
import base64
import shutil
import math
from dateutil import tz
from datetime import datetime
import requests


def parse_file_name(file_name, prefix):
    filename = file_name.split(prefix)[-1]
    return filename


def convert_size(size_bytes):
    if size_bytes == 0:
        return "0B"
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)
    return "%s %s" % (s, size_name[i])


def format_datetime(date_obj):
    from_zone = tz.tzutc()
    to_zone = tz.tzlocal()
    utc = date_obj.replace(tzinfo=from_zone)
    # Convert time zone
    central = utc.astimezone(to_zone)
    return str(central.strftime("%b %d, %Y %H:%M"))


def fetch_dir_file_paths(path, exclude_file):
    list_of_paths = list()
    for root, dirs, files in os.walk(path):
        for file in files:
            data_paths = os.path.join(root, file)
            data_paths = data_paths.split(path)[-1]
            print(data_paths)
            list_of_paths.append(data_paths)
    list_of_paths.remove(exclude_file.split(path)[-1])
    return list_of_paths


def convert_into_bytes(size_string):
    units = {"KB": 1024, "MB": 1024 * 1024, "GB": 1024 * 1024 * 1024}
    for unit in units:
        if unit in size_string:
            index = size_string.find(unit)
            size = size_string[:index].strip()
            unit = size_string[index:].strip()
            return float(size) * units[unit]

    # unit is B
    return float(size_string[:-1])



# -*- coding: utf-8 -*-
import json
from .utils import NumpyArrayEncoder


def load_model(model_path):
    import cloudpickle

    return cloudpickle.load(open(model_path, "rb"))


def dump_model(model, path):
    import cloudpickle

    cloudpickle.dump(model, open(path, "wb"))


def get_model_structure(model_obj):
    ml_model_class = model_obj.__dict__
    ml_model_class["class"] = str(model_obj.__class__)[8:-2:]
    return json.dumps(ml_model_class, cls=NumpyArrayEncoder)



# -*- coding: utf-8 -*-
import base64
import json
import requests
from urllib.parse import quote, quote_plus
import tempfile
import os
import shutil

from .constants import Notebook, MosaicVersionControl, Headers


def get_repo(project_id=None, email_address=None, mosaicid=None, username=None, repo_id=None):
    """
    Common function for getting enabled git repo for the project

    Args:
        :param project_id:
        :param email_address:
        :param mosaicid:
        :param username:
        :param repo_id:
    """
    try:
        nb_url = Notebook.url + \
                 (Notebook.get_repo_by_id_api+repo_id if repo_id else Notebook.git_access_api)
        headers = {
            'x-auth-email': email_address,
            'x-auth-userid': mosaicid,
            'x-auth-username': username,
            'x-project-id': project_id,
        }
        enabled_repo = requests.get(nb_url, headers=headers)
        if enabled_repo and enabled_repo.status_code == 200:
            repo_details = enabled_repo.json()
            repo_url = repo_details["repo_url"]
            url_parts = repo_url.split("//")
            repo_type = repo_details["repo_type"]
            # password = replace_special_chars_with_ascii(str(git_access_token))
            password = quote(repo_details["password"], safe='')

            remote_url = "{0}//{1}:{2}@{3}".format(
                url_parts[0],
                repo_details["username"],
                password,
                url_parts[1],
            )
            repository = {"url": remote_url,
                          "base_folder": repo_details["base_folder"],
                          "branch": repo_details["branch"],
                          "repo_url": repo_url,
                          "repo_type": repo_type,
                          "proxy_details": repo_details["proxy_details"]
                          }
        else:
            repository = None
    except Exception as ex:
        repository = None

    return repository


def validate_repo(remote_url, git_branch, file_path):
    """
    Common function for validating a git repo

    Args:
        :param remote_url:
        :param git_branch:
        :param file_path:
    """
    try:
        git_temp_dir = tempfile.mkdtemp()

        Repo.clone_from(remote_url, git_temp_dir)
        repo = Repo(git_temp_dir)

        branch_exist = True
        try:
            repo.git.checkout(git_branch)
        except Exception as ex:
            branch_exist = False

        path_exist = os.path.exists(os.path.join(git_temp_dir, file_path))
        shutil.rmtree(git_temp_dir)
        if branch_exist and path_exist:
            return "Success", True
        if branch_exist is False:
            return "Invalid branch name. Please provide a valid branch name.", False
        if path_exist is False:
            return "Invalid base path. Please provide a valid base path.", False

    except Exception as ex:
        return "Invalid details. Please provide the correct information.", False


def create_remote_url(data):
    """
    Creates git clone url in format http(s)://user:token@remoteurl/repo.git
    :param data:
    :return:
    """
    url_parts = data['repo_url'].split("//")
    password = quote(data["password"], safe='')
    remote_url = "{0}//{1}:{2}@{3}".format(
        url_parts[0],
        data["username"],
        password,
        url_parts[1],
    )
    return remote_url


def encode_b64(input_str):
    """
    Method encodes a string to base64 encoding
    :param input_str:
    :return:
    """
    return base64.b64encode(input_str.encode("UTF-8")).decode("UTF-8") if input_str else ""


def decode_b64(input_str):
    """
    Method decodes the base64 encoding
    :param input_str:
    :return:
    """
    return base64.b64decode(input_str.encode("UTF-8")).decode("UTF-8") if input_str else ""


def encode_password(repo_object, skip_access_cat=None):
    """
    Encode a password field from repo object to base64
    :param repo_object: The repository Object,
    :param skip_access_cat: skip Encoding for given access category
    :return:
    """
    return _encode_decode_password(repo_object, skip_access_cat, is_encode=True)


def decode_password(repo_object, skip_access_cat=None):
    """
    Decode a password field from repo object from base64 to plain text
    :param repo_object: The repository Object
    :param skip_access_cat: skip decoding for given access category
    :return:
    """
    return _encode_decode_password(repo_object, skip_access_cat, is_encode=False)


def _encode_decode_password(repo_object, skip_access_cat, is_encode=False):
    """
    Common function for encode and decode
    :param repo_object:
    :param skip_access_cat:skip encode/decode for given access category
    :param is_encode:
    :return:
    """
    skip_access_cat = skip_access_cat.upper() if skip_access_cat else skip_access_cat
    if repo_object and repo_object.get("access_category", "PUBLIC").upper() != skip_access_cat:
        secure_flag = str(repo_object.get("secure_flag", "true")).lower() == "true"
        if not secure_flag:
            passwd = repo_object.get("password", "")
            repo_object["password"] = encode_b64(passwd) if is_encode else decode_b64(passwd)
    return repo_object
  
    
def state_of_code(project_id, version_id, repo_id="", branch="", headers=None):
    """Method to maintain state of code"""
    url = MosaicVersionControl.url + MosaicVersionControl.git_tag.format(
        repo=project_id, version_id=version_id, repo_id=repo_id, branch=branch)
    if not headers:
    	headers = get_headers()
    response = requests.post(url, headers=headers)
    response.raise_for_status()
    return response


def get_headers():
    """
    Create headers
    :return:
    """
    return {
        Headers.x_userid: os.getenv("userId"),
        Headers.x_project_id: os.getenv("PROJECT_ID"),
        Headers.x_username: os.getenv("userId"),
        Headers.x_email_id: os.getenv("userId")
    }


def extract_proxy_values(proxy_details : tuple):
    """
    This function extracts the proxy details and retuns a dictionary
    """
    if type(proxy_details) == str:
        proxy_details = json.loads(proxy_details)
    proxy_ip = proxy_details.get("IPaddress")
    verify_ssl = str(proxy_details.get("SSLVerify", "")).lower()
    proxy_username = proxy_details.get('UsernameOfProxy', None)
    proxy_password = proxy_details.get('ProxyPassword', None)
    proxy_protocol = proxy_details.get("Protocol", "http")
    if proxy_password:
        proxy_password = quote_plus(proxy_password)
    return (proxy_ip, verify_ssl, proxy_username, proxy_password, proxy_protocol)



'''
POC: Testing Scenarios Not To Be Considered For Production
'''

import requests
import logging
import pandas as pd
import numpy as np

logging.basicConfig(level=logging.DEBUG)
from abc import ABC, abstractmethod
from typing import Tuple, Union, List, Iterable, Any
from .constants import ScoreLabel
from .constants import ValidationBoxLabel as VBLabel
from .validators import RequestValidatorMixin, PayloadSchemaValidatorMixin

from .response import ScoreResponse
from .encoder import ScoreResponseEncoder
from collections import OrderedDict
from IPython.display import display

try:
    pd.set_option('display.max_colwidth', -1)
except Exception:
    pd.set_option('display.max_colwidth', None)


class ScoreBase(RequestValidatorMixin,
                PayloadSchemaValidatorMixin,
                ABC):
    each_request_processed_schema: OrderedDict = NotImplementedError
    each_pre_processed_schema: OrderedDict = NotImplementedError

    def __init__(self):
        self.meta = self.Meta()
        self.score_request = None
        self.score_response_collection = []
        self.n_input = None
        self.all_request_processed_payload = None
        self.all_clean_payload = None

        self.validation_box = pd.DataFrame(
            {
                "VALIDATION": pd.Series(VBLabel.VALIDATION,
                                        dtype=str),
                "COMPONENT": pd.Series(VBLabel.COMPONENT,
                                       dtype=str),
                "PASSED": pd.Series(VBLabel.PASSED,
                                    dtype=bool),
                "SKIPPED": pd.Series(VBLabel.SKIPPED,
                                     dtype=bool),
            }
        )

    @abstractmethod
    def request_processing_fn(self,
                              request,
                              *args,
                              **kwargs) -> Tuple[int, Union[List[np.ndarray], List[Any]]]:
        """
        Segregates Payload From Request Object
        Request Object -> (N,  List[Input data, ..])

        Input data could be:
            - List[ [Feature_Value1, .., Feature_ValueN], [...] ]
            - List[numpy.array(), numpy.array(), ...]
            - List[tf.Tensor, tf.Tensor, tf.Tensor, ...]
            - List[ SingleSample, SingleSample]

        A Numpy array (or array-like), or a list of arrays (in case the model has multiple inputs).
        A TensorFlow tensor, or a list of tensors (in case the model has multiple inputs).
        A dict mapping input names to the corresponding array/tensors, if the model has named inputs.

        :return: (n_inputs, payload's)

        Warnings:

        1. Do not reshape your final output for single sample here(array.reshape(1, -1)), do it in prediction_fn.
           Else payloads will be invalidated for extraction at raw and extraction level.
        """
        pass

    @abstractmethod
    def pre_processing_fn(self,
                          each_payload: Union[np.ndarray, Any]) -> Union[np.ndarray, Any]:
        """
        Accepts Single Payload At a time and applies pre_processing_step before prediction

        :param each_payload: Single Payload
        :return: Preprocessed Payload
        """
        pass

    @abstractmethod
    def prediction_fn(self,
                      model: Any,
                      pre_processed_input: Union[np.ndarray, Any]
                      ):
        """
        Does the main prediction on pre_processed_input using supplied model.

        :param model: Supported Model
        :param pre_processed_input: Single Preprocessed Payload
        :return: Prediction Value From the model

        Recommended Notes:
        - Numpy:
            - Reshape your data array.reshape(1, -1) here before predictions as it contains a single sample.
        """
        pass

    def score(self,
              model,
              input_request,
              dry_run=False):
        self.reset()

        if self.can_run_pre_call_hooks():
            for each_hooks in self.meta.pre_call_hooks:
                each_hooks()

        _request_processed_fn_op = self.request_processing_fn(input_request)
        self.validate_request(_request_processed_fn_op)

        if dry_run:
            display(self.validation_box[:3])
            print(ScoreLabel.Note, "\n")

        if not self.validation_box["PASSED"].all():
            return {"error": "Validation Failed. Please run with dry_run=True to see all exceptions."}

        self.n_input, self.all_request_processed_payload = _request_processed_fn_op

        for each_request in self.get_processed_request_payload():
            _rp_payload_validated = False
            _clean_payload_validated = False

            if self.validate_with_schema_payload(self.each_request_processed_schema,
                                                 each_request):
                _rp_payload_validated = True

            _clean_payload = self.pre_processing_fn(each_request)
            if self.validate_with_schema_payload(self.each_pre_processed_schema,
                                                 _clean_payload):
                _clean_payload_validated = True

            _score_response = self.prediction_fn(model, _clean_payload)

            _response = ScoreResponse(score_request=input_request.json,
                                      request_processed_payload=each_request,
                                      rp_payload_validated=_rp_payload_validated,
                                      clean_payload=_clean_payload,
                                      clean_payload_validated=_clean_payload_validated,
                                      score_response=_score_response)

            self.score_response_collection.append(_response)

        if self.can_run_post_call_hooks():
            for each_hooks in self.meta.post_call_hooks:
                each_hooks()

        return self.score_response_collection

    def reset(self):
        self.score_request = None
        self.score_response_collection = []
        self.n_input = None
        self.all_request_processed_payload = None
        self.all_clean_payload = None

    def get_processed_request_payload(self) -> Iterable:
        """

        :return:  Iterables of Inputs
        """
        if isinstance(self.all_request_processed_payload, list):
            return iter(self.all_request_processed_payload)
        else:
            return iter([self.all_request_processed_payload])

    def get_clean_payload(self) -> Iterable:
        """

        :return:  Iterables of Inputs
        """

        if isinstance(self.all_clean_payload, list):
            return iter(self.all_clean_payload)
        else:
            return iter([self.all_clean_payload])

    def can_run_pre_call_hooks(self):
        """
        :return: True if this score has a `pre_call_hooks` attributes defined for with list of Callables,
        False otherwise.
        """
        return hasattr(self.__class__.Meta, "pre_call_hooks")

    def can_run_post_call_hooks(self):
        """
        :return: True if this score has a `post_call_hooks` attributes defined for with list of Callables,
        False otherwise.
        """
        return hasattr(self.__class__.Meta, "post_call_hooks")

    @staticmethod
    def download_template():
        from os.path import abspath, join, dirname
        get_cwd = dirname(abspath(__file__))
        with open(join(get_cwd, "templates.py"), 'r') as template_example:
            content = template_example.readlines()
            content = list(map(lambda each_line: each_line.strip("\n"), content))
            for lines in content:
                print(lines)

    class Meta:
        pass



# -*- coding: utf-8 -*-
from base64 import b64decode, b64encode
import numpy as np
import pandas as pd

def base64_encode(data: str) -> str:
    """ Encodes input string in base64 format """
    data = data.encode("utf-8")
    data = b64encode(data)
    return data.decode("utf-8")


def fix_padding(encoded_string: str) -> str:
    """ method to correct padding for base64 encoding """
    required_padding = len(encoded_string) % 4
    return encoded_string + ("=" * required_padding)


def base64_decode(data: str) -> str:
    """ decodes input string from base64 format"""
    data = data.encode("utf-8")
    data = b64decode(data)
    return data.decode("utf-8")


class encoder:
    """Encode categorical features."""
    def __init__(self, cat_cols, encoder=None):
        if encoder:
            self.encoder = encoder
        else:
            from sklearn.preprocessing import OneHotEncoder
            self.encoder = OneHotEncoder(categories="auto")
        self.cat_cols = cat_cols

    def encode(self, features, fit=False):
        """
        Encode the categorical features and join return merged data.

        Args:
            features: numpy array of features
            fit:    : boolean default False
                      If yes, it performs fit_transform operation
                      else, it performs only transform operation
        Returns:
            Numpy Array with the non categorical features combined
            with the encoded categorical features.
        """
        if not self.cat_cols:
            return None, features

        non_cat_cols = (
            sorted(set(range(features.shape[1])) - set(self.cat_cols)))
        if fit:
            cat_data = self.encoder.fit_transform(features[:, self.cat_cols])
        else:
            cat_data = self.encoder.transform(features[:, self.cat_cols])
        features = np.hstack(
            (features[:, non_cat_cols].astype('float64'), cat_data.toarray()))
        return non_cat_cols, features

    def get_encoder(self):
        return self.encoder


def find_categorical(arry, unique_num_count=10):
    """
        Get categorical and non categorical columns.

        It will treat all non numeric fields for encoding along with
        numeric data having less than unique_num_count unique values.

        """
    df = pd.DataFrame(arry)
    num_cols = set(df._get_numeric_data().columns)
    df = df.apply(lambda x: pd.to_numeric(x, errors="ignore"))
    num_cols = set(df._get_numeric_data().columns)
    cat_cols = set(df.columns) - num_cols

    for col in num_cols:
        if df[col].nunique() <= unique_num_count:
            cat_cols.add(col)

    non_cat_cols = sorted(set(df.columns) - cat_cols)
    cat_cols = sorted(cat_cols)
    return cat_cols, non_cat_cols



import datetime
from uuid import uuid4
from .constants import ScoreResponseCollectionLabel as SCRLabel


class ScoreResponse(object):
    purpose = "<<ScoreResponse>>"
    fields = SCRLabel.SCORE_RESPONSE_FIELDS

    def __repr__(self):
        return self.purpose

    def __init__(self,
                 uuid=str(uuid4()),
                 timestamp=datetime.datetime.now(),
                 score_request=None,
                 request_processed_payload=None,
                 rp_payload_validated=False,
                 clean_payload=None,
                 clean_payload_validated=False,
                 score_response=None,
                 ):

        self._uuid = uuid
        self.timestamp = timestamp
        self.score_request = score_request
        self.request_processed_payload = request_processed_payload
        self.rp_payload_validated = rp_payload_validated
        self.clean_payload = clean_payload
        self.clean_payload_validated = clean_payload_validated
        self.score_response = score_response

    @property
    def uuid(self) -> str:
        return self._uuid

    @uuid.setter
    def uuid(self, uuid: str) -> None:
        if not isinstance(uuid, str):
            raise ValueError("UUID Field Must Be String")
        self._uuid = uuid

    def to_dict(self):
        tmp_ = dict.fromkeys(self.fields)
        for k, v in tmp_.items():
            tmp_[k] = getattr(self, k)
        return tmp_



from datetime import  datetime


class ScoreLabel:
    Note= "Fields Marked Asterisk (*) Can Be Validated On Proper Input"
    Request_Return_Type = "Return Type Must Be Tuple (n_input, payloads)"
    Request_Return_Shape = "Tuple Must Be of length Two (n_input, payloads)"
    Request_Many_Payload_DType = "* if n_input > 1 payload type must be List (n_input, [np.ndarray, tf.Tensor, etc])"
    ERP_Schema_Dtype = "variable each_request_processed_schema must be of OrderedDict"
    Correct_Mapping_With_Schema = "Each Payload Is Successfully Validated With each_request_processed_schema" \
                                  " after request_processing_fn"


class ValidationBoxLabel:
    VALIDATION = [ScoreLabel.Request_Return_Type,
                  ScoreLabel.Request_Return_Shape,
                  ScoreLabel.Request_Many_Payload_DType,
                 ]

    COMPONENT = ["request_processing_fn",
                 "request_processing_fn",
                 "request_processing_fn",
                 ]

    PASSED = [False,
              False,
              True,
              ]

    SKIPPED = [True for i in range(3)]


class ScoreResponseLabel:
    TIMESTAMP_FORMAT = "%m/%d/%Y, %H:%M:%S"


class ScoreResponseCollectionLabel:
    SCORE_RESPONSE_FIELDS = ["_uuid",
                             "timestamp",
                             "score_request",
                             "request_processed_payload",
                             "rp_payload_validated",
                             "clean_payload",
                             "clean_payload_validated",
                             "score_response"]

    SCORE_RESPONSE_DTYPES = ['str',
                             'datetime',
                             'object',
                             'object',
                             'bool',
                             'object',
                             'bool',
                             'object'
                             ]


class Minio:
    url = "MINIO_URL"
    bucket = "MINIO_BUCKET"
    access_key = "MINIO_ACCESS_KEY"
    secret_key = "MINIO_SECRET_KEY"



from mosaic_utils.ai.encoding_utils import find_categorical

"""Target Encoder"""
import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator
from category_encoders.ordinal import OrdinalEncoder
import category_encoders.utils as util
'''
def  actual_fit_transform(X,y =None):
    enc = TargetEncoder(cols=[cat_cols]).fit_transform(X,y)
    return enc

class TargetEncoder:
    def __init__(self, verbose=0, cols=None, drop_invariant=False, return_df=True, handle_missing='value',
                     handle_unknown='value', min_samples_leaf=1, smoothing=1.0):
        self.return_df = return_df
        self.drop_invariant = drop_invariant
        self.drop_cols = []
        self.verbose = verbose
        self.cols = cols
        self.ordinal_encoder = None
        self.min_samples_leaf = min_samples_leaf
        self.smoothing = float(smoothing)  # Make smoothing a float so that python 2 does not treat as integer division
        self._dim = None
        self.mapping = None
        self.handle_unknown = handle_unknown
        self.handle_missing = handle_missing
        self._mean = None
        self.feature_names = None

    def fit(self, X, y=None):
        """Fi OneHotEncoder to X.

        Parameters
        ----------
        X : array-like, shape=(n_samples, n_feature)
            Input array of type int.

        Returns
        -------
        self
        """
        self.fit_transform(X)
        return selft

    def fit_transform(self, X, y=None):
        self.cat_cols, self.non_cat_cols = find_categorical(X,10)
        X_transformed = actual_fit_transform(self.cat_cols,y)
        return X_transformed





'''

class TargetEncoder(BaseEstimator, util.TransformerWithTargetMixin):
    """Target encoding for categorical features.

    For the case of categorical target: features are replaced with a blend of posterior probability of the target
    given particular categorical value and the prior probability of the target over all the training data.

    For the case of continuous target: features are replaced with a blend of the expected value of the target
    given particular categorical value and the expected value of the target over all the training data.

    Parameters
    ----------

    verbose: int
        integer indicating verbosity of the output. 0 for none.
    cols: list
        a list of columns to encode, if None, all string columns will be encoded.
    drop_invariant: bool
        boolean for whether or not to drop columns with 0 variance.
    return_df: bool
        boolean for whether to return a pandas DataFrame from transform (otherwise it will be a numpy array).
    handle_missing: str
        options are 'error', 'return_nan'  and 'value', defaults to 'value', which returns the target mean.
    handle_unknown: str
        options are 'error', 'return_nan' and 'value', defaults to 'value', which returns the target mean.
    min_samples_leaf: int
        minimum samples to take category average into account.
    smoothing: float
        smoothing effect to balance categorical average vs prior. Higher value means stronger regularization.
        The value must be strictly bigger than 0.

    Example
    -------
    >>> from category_encoders import *
    >>> import pandas as pd
    >>> from sklearn.datasets import load_boston
    >>> bunch = load_boston()
    >>> y = bunch.target
    >>> X = pd.DataFrame(bunch.data, columns=bunch.feature_names)
    >>> enc = TargetEncoder(cols=['CHAS', 'RAD']).fit(X, y)
    >>> numeric_dataset = enc.transform(X)
    >>> print(numeric_dataset.info())
    <class 'pandas.core.frame.DataFrame'>
    RangeIndex: 506 entries, 0 to 505
    Data columns (total 13 columns):
    CRIM       506 non-null float64
    ZN         506 non-null float64
    INDUS      506 non-null float64
    CHAS       506 non-null float64
    NOX        506 non-null float64
    RM         506 non-null float64
    AGE        506 non-null float64
    DIS        506 non-null float64
    RAD        506 non-null float64
    TAX        506 non-null float64
    PTRATIO    506 non-null float64
    B          506 non-null float64
    LSTAT      506 non-null float64
    dtypes: float64(13)
    memory usage: 51.5 KB
    None

    References
    ----------

    .. [1] A Preprocessing Scheme for High-Cardinality Categorical Attributes in Classification and Prediction Problems, from
    https://dl.acm.org/citation.cfm?id=507538

    """

    def __init__(self, verbose=0, cols=None, drop_invariant=False, return_df=True, handle_missing='value',
                     handle_unknown='value', min_samples_leaf=1, smoothing=1.0):
        self.return_df = return_df
        self.drop_invariant = drop_invariant
        self.drop_cols = []
        self.verbose = verbose
        self.cols = cols
        self.ordinal_encoder = None
        self.min_samples_leaf = min_samples_leaf
        self.smoothing = float(smoothing)  # Make smoothing a float so that python 2 does not treat as integer division
        self._dim = None
        self.mapping = None
        self.handle_unknown = handle_unknown
        self.handle_missing = handle_missing
        self._mean = None
        self.feature_names = None


    def fit(self, X, y, **kwargs):
        """Fit encoder according to X and y.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number of samples
            and n_features is the number of features.
        y : array-like, shape = [n_samples]
            Target values.

        Returns
        -------
        self : encoder
            Returns self.

        """

        # unite the input into pandas types
        X = util.convert_input(X)
        y = util.convert_input_vector(y, X.index)

        if X.shape[0] != y.shape[0]:
            raise ValueError("The length of X is " + str(X.shape[0]) + " but length of y is " + str(y.shape[0]) + ".")
           
        self._dim = X.shape[1]
        
        self.cols,_ = find_categorical(X,10) 
        # if columns aren't passed, just use every string column
        if self.cols is None:
            self.cols = util.get_obj_cols(X)
        else:
            self.cols = util.convert_cols_to_list(self.cols)

        if self.handle_missing == 'error':
            if X[self.cols].isnull().any().any():
                raise ValueError('Columns to be encoded can not contain null')

        self.ordinal_encoder = OrdinalEncoder(
            verbose=self.verbose,
            cols=self.cols,
            handle_unknown='value',
            handle_missing='value'
        )
        self.ordinal_encoder = self.ordinal_encoder.fit(X)
        X_ordinal = self.ordinal_encoder.transform(X)
        
        self.mapping = self.fit_target_encoding(X_ordinal, y)
        
        X_temp = self.transform(X, override_return_df=True)
        self.feature_names = list(X_temp.columns)

        if self.drop_invariant:
            self.drop_cols = []
            X_temp = self.transform(X)
            generated_cols = util.get_generated_cols(X, X_temp, self.cols)
            self.drop_cols = [x for x in generated_cols if X_temp[x].var() <= 10e-5]
            try:
                [self.feature_names.remove(x) for x in self.drop_cols]
            except KeyError as e:
                if self.verbose > 0:
                    print("Could not remove column from feature names."
                    "Not found in generated cols.\n{}".format(e))

        return self


    def fit_target_encoding(self, X, y):
        mapping = {}

        for switch in self.ordinal_encoder.category_mapping:
            col = switch.get('col')
            values = switch.get('mapping')

            prior = self._mean = y.mean()

            stats = y.groupby(X[col]).agg(['count', 'mean'])

            smoove = 1 / (1 + np.exp(-(stats['count'] - self.min_samples_leaf) / self.smoothing))
            smoothing = prior * (1 - smoove) + stats['mean'] * smoove
            smoothing[stats['count'] == 1] = prior

            if self.handle_unknown == 'return_nan':
                smoothing.loc[-1] = np.nan
            elif self.handle_unknown == 'value':
                smoothing.loc[-1] = prior

            if self.handle_missing == 'return_nan':
                smoothing.loc[values.loc[np.nan]] = np.nan
            elif self.handle_missing == 'value':
                smoothing.loc[-2] = prior

            mapping[col] = smoothing
            

        return mapping

    def transform(self, X, y=None, override_return_df=False):
        """Perform the transformation to new categorical data.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
        y : array-like, shape = [n_samples] when transform by leave one out
            None, when transform without target info (such as transform test set)
            
        Returns
        -------
        p : array, shape = [n_samples, n_numeric + N]
            Transformed values with encoding applied.

        """

        if self.handle_missing == 'error':
            if X[self.cols].isnull().any().any():
                raise ValueError('Columns to be encoded can not contain null')

        if self._dim is None:
            raise ValueError('Must train encoder before it can be used to transform data.')

        # unite the input into pandas types
        X = util.convert_input(X)

        # then make sure that it is the right size
        if X.shape[1] != self._dim:
            raise ValueError('Unexpected input dimension %d, expected %d' % (X.shape[1], self._dim,))

        # if we are encoding the training data, we have to check the target
        if y is not None:
            y = util.convert_input_vector(y, X.index)
            if X.shape[0] != y.shape[0]:
                raise ValueError("The length of X is " + str(X.shape[0]) + " but length of y is " + str(y.shape[0]) + ".")

        if not list(self.cols):
            return X

        X = self.ordinal_encoder.transform(X)

        if self.handle_unknown == 'error':
            if X[self.cols].isin([-1]).any().any():
                raise ValueError('Unexpected categories found in dataframe')

        X = self.target_encode(X)

        if self.drop_invariant:
            for col in self.drop_cols:
                X.drop(col, 1, inplace=True)

        if self.return_df or override_return_df:
            return X
        else:
            return X.values


    def target_encode(self, X_in):
        X = X_in.copy(deep=True)

        for col in self.cols:
            X[col] = X[col].map(self.mapping[col])

        return X

    def get_feature_names(self):
        """
        Returns the names of all transformed / added columns.

        Returns
        -------
        feature_names: list
            A list with all feature names transformed or added.
            Note: potentially dropped features are not included!

        """

        if not isinstance(self.feature_names, list):
            raise ValueError('Must fit data first. Affected feature names are not known before.')
        else:
            return self.feature_names




from .constants import *
import requests
from flask import Response


def generate_headers(userid=None, email=None, username=None, project_id=None, x_request_id=None, access_control=None):
    """
    Common function for generating headers based on given parameters

    Args:
        :param access_control:
        :param x_request_id:
        :param project_id:
        :param username:
        :param email:
        :param userid:
    """
    # assigning headers for api
    final_headers = {
        Headers.x_auth_userid: userid,
        Headers.x_auth_email: email,
        Headers.x_auth_username: username,
        Headers.x_project_id: project_id,
        Headers.x_request_id: x_request_id,
        Headers.access_control_allow_origin: access_control,
        Headers.access_control_allow_methods: access_control,
        Headers.access_control_allow_headers: access_control
    }
    return final_headers


def check_project_access(console_url, userid, email, username, project_id, access=None):
    """
      Common function for checking project access

      Args:
          :param console_url:
          :param project_id:
          :param username:
          :param email:
          :param userid:
          :param access
      """
    headers = generate_headers(userid, email, username, project_id)
    project_access_url = f"{console_url}/secured/api/project/v1/access"
    response = requests.get(project_access_url, headers=headers)
    if response.status_code == 404 or response.status_code == 401:
        return Response("Access denied for given project ", status=403)
    if response.status_code == 500:
        if "not authorised" in response.json()["message"]:
            raise ValueError("Access denied for this project.")
        if "not found" in response.json()["message"]:
            raise ValueError("Project you are trying to access has been deleted.")
    if response.status_code == 200 and access and response.json()["accessType"] not in ["OWNER", "CONTRIBUTOR", "VALIDATOR"]:
        raise ValueError("Access denied for this project.")



# -*- coding: utf-8 -*-
"""validations.py: Defines all validations."""

__author__ = "Shivam Chaurasia (10670364)"
__description__ = "Defines all validations by implementing IValidation Interface"
__version__ = "0.1"

from abc import ABC, abstractmethod
from typing import Dict, Tuple
import logging
import re


class IValidation(ABC):
    """
    Abstract Interface For Implementing Validations
    """
    purpose = NotImplementedError
    _fields = NotImplementedError

    def __init__(self,
                 fields: Dict = None):
        self._fields = dict()
        if fields:
            self._fields.update(fields)

    @abstractmethod
    def check(self) -> Tuple[bool, str]:
        """Interface Implementing Actual Validation"""
        pass

    def __repr__(self):
        return self.purpose


class MandatoryFields(IValidation):
    """
    Validates provided value must exist.

    Usage:
    MandatoryFields(fields: Dict=mandatory_fields)
    :example
        obj = MandatoryFields(fields = {"test_name", 1000})
        assert obj.check() == True
    """
    purpose = "Validation/MandatoryValues"

    def check(self):
        for field_name, value in self._fields.items():
            if value is None or value == '' or str(value).strip() == '':
                return False, f"{field_name} cannot be empty, blank or whitespace"
        else:
            message_ = "PASS:: Mandatory Validation :: {}".format(
                self._fields.keys()
            )
            logging.info(message_)
        return True, "Mandatory_Fields_Validated"


class ANumericUScoresExclusive(IValidation):
    """
    Validates provided value must be AlphaNumeric(A-Z,a-z,0-9) and underscores.
    Starting and Ending underscore are not allowed.

    Usage:
    ANumericUScoresExclusive(fields: Dict=mandatory_fields)
    :example
        obj = ANumericUScoresExclusive(fields = {"test_name", "Test_001"})
        assert obj.check() == True
    """
    purpose = "Validation/AlphaNumericAndUScoreExclusive"

    def check(self):
        pattern_alpnumeric = re.compile("^\w+$")
        for field_name, value in self._fields.items():
            if value.startswith("_") or value.endswith("_"):
                return False, f"{field_name} cannot start or ends with underscores"
            if not re.match(pattern_alpnumeric, value):
                return False, f"{field_name} can only contain alphanumeric and underscore"
        else:
            message_ = "PASS:: AlphaNumeric Validation :: {}".format(
                self._fields.keys()
            )
            logging.info(message_)
        return True, "Alphanumeric_Underscore_Exclusive_Fields_Validated"


class IfPresentTypeCheck(IValidation):
    """
    Validates if parent value exist check subfield instance datatypes.

    Usage:
    IfPresentTypeCheck(fields: Dict={"parent_field": (parent_field_val, (dtype_1, dtype_n) )})
    :example
        obj = IfPresentTypeCheck(fields: Dict={"test_parent_field": (True, (int, bool) )})
        assert obj.check() == True
    """
    purpose = "Validation/IfPresentTypeCheck"

    def check(self):
        for field_name, (if_present_value, tuple_of_dtypes) in self._fields.items():
            if if_present_value is not None:
                if not isinstance(if_present_value, tuple_of_dtypes):
                    return False, f"{field_name} must be of type - f{tuple_of_dtypes}"
        else:
            message_ = "PASS:: {} :: {}".format(
                repr(self),
                self._fields.keys()
            )
            logging.info(message_)
        return True, repr(self)


class IfPresentSubfieldMustExist(IValidation):
    """
    Validates if parent value exist check subfield must exist.

    Usage:
    IfPresentSubfieldMustExist(fields: Dict=
        {"parent_field": (parent_field_val,
                            {"subfield_n": subfield_n_value}
                        )
        })
    :example
        obj = IfPresentSubfieldMustExist(fields: Dict={"parent_field": (True,
                                                                            {"test_sub1": "non_empty_val",
                                                                             "test_sub2": "non_empty_val",
                                                                             }
                        )
        })
        assert obj.check() == True
    """
    purpose = "Validation/IfPresentSubfieldMustExist"

    def check(self):
        for field_name, (if_present_value, dict_of_sub_field) in self._fields.items():
            if if_present_value is not None and if_present_value is not False:
                for each_value in  dict_of_sub_field.values():
                    if each_value is None or each_value is '':
                        message_ = "If {field_name} present -> {fields_required} Mandatory"
                        message_ = message_.format(field_name=field_name,
                                                   fields_required=list(dict_of_sub_field.keys())
                                               )
                        return False, message_
        else:
            message_ = "PASS:: {} :: {}".format(
                repr(self),
                self._fields.keys()
            )
            logging.info(message_)
        return True, repr(self)


import pandas as pd
import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from patsy.contrasts import Helmert
from category_encoders.ordinal import OrdinalEncoder
import category_encoders.utils as util
from mosaic_utils.ai.encoding_utils import find_categorical
class HelmertEncoder(BaseEstimator, TransformerMixin):
    """Helmert contrast coding for encoding categorical features.

    Parameters
    ----------

    verbose: int
        integer indicating verbosity of the output. 0 for none.
    cols: list
        a list of columns to encode, if None, all string columns will be encoded.
    drop_invariant: bool
        boolean for whether or not to drop columns with 0 variance.
    return_df: bool
        boolean for whether to return a pandas DataFrame from transform (otherwise it will be a numpy array).
    handle_unknown: str
        options are 'error', 'return_nan', 'value', and 'indicator'. The default is 'value'. Warning: if indicator is used,
        an extra column will be added in if the transform matrix has unknown categories.  This can cause
        unexpected changes in dimension in some cases.
    handle_missing: str
        options are 'error', 'return_nan', 'value', and 'indicator'. The default is 'value'. Warning: if indicator is used,
        an extra column will be added in if the transform matrix has nan values.  This can cause
        unexpected changes in dimension in some cases.

    Example
    -------
    >>> from category_encoders import *
    >>> import pandas as pd
    >>> from sklearn.datasets import load_boston
    >>> bunch = load_boston()
    >>> y = bunch.target
    >>> X = pd.DataFrame(bunch.data, columns=bunch.feature_names)
    >>> enc = HelmertEncoder(cols=['CHAS', 'RAD'], handle_unknown='value', handle_missing='value').fit(X, y)
    >>> numeric_dataset = enc.transform(X)
    >>> print(numeric_dataset.info())
    <class 'pandas.core.frame.DataFrame'>
    RangeIndex: 506 entries, 0 to 505
    Data columns (total 21 columns):
    intercept    506 non-null int64
    CRIM         506 non-null float64
    ZN           506 non-null float64
    INDUS        506 non-null float64
    CHAS_0       506 non-null float64
    NOX          506 non-null float64
    RM           506 non-null float64
    AGE          506 non-null float64
    DIS          506 non-null float64
    RAD_0        506 non-null float64
    RAD_1        506 non-null float64
    RAD_2        506 non-null float64
    RAD_3        506 non-null float64
    RAD_4        506 non-null float64
    RAD_5        506 non-null float64
    RAD_6        506 non-null float64
    RAD_7        506 non-null float64
    TAX          506 non-null float64
    PTRATIO      506 non-null float64
    B            506 non-null float64
    LSTAT        506 non-null float64
    dtypes: float64(20), int64(1)
    memory usage: 83.1 KB
    None

    References
    ----------

    .. [1] Contrast Coding Systems for Categorical Variables, from
    https://stats.idre.ucla.edu/r/library/r-library-contrast-coding-systems-for-categorical-variables/

    .. [2] Gregory Carey (2003). Coding Categorical Variables, from
    http://psych.colorado.edu/~carey/Courses/PSYC5741/handouts/Coding%20Categorical%20Variables%202006-03-03.pdf

    """

    def __init__(self, verbose=0, cols=None, mapping=None, drop_invariant=False, return_df=True,
                 handle_unknown='value', handle_missing='value'):
        self.return_df = return_df
        self.drop_invariant = drop_invariant
        self.drop_cols = []
        self.verbose = verbose
        self.mapping = mapping
        self.handle_unknown = handle_unknown
        self.handle_missing = handle_missing
        self.cols = cols
        self.ordinal_encoder = None
        self._dim = None
        self.feature_names = None
    
    def fit(self, X, y=None, **kwargs):
        """Fit encoder according to X and y.

        Parameters
        ----------

        X : array-like, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number of samples
            and n_features is the number of features.
        y : array-like, shape = [n_samples]
            Target values.

        Returns
        -------

        self : encoder
            Returns self.

        """
        
        # first check the type
        X = util.convert_input(X)
        
        self._dim = X.shape[1]
        
        self.cols,_ = find_categorical(X,10)
        # if columns aren't passed, just use every string column
        if self.cols is None:
            self.cols = util.get_obj_cols(X)
        else:
            self.cols = util.convert_cols_to_list(self.cols)

        if self.handle_missing == 'error':
            if X[self.cols].isnull().any().any():
                raise ValueError('Columns to be encoded can not contain null')

        self.ordinal_encoder = OrdinalEncoder(
            verbose=self.verbose,
            cols=self.cols,
            handle_unknown='value',
            handle_missing='value'
        )
        self.ordinal_encoder = self.ordinal_encoder.fit(X)
       
        ordinal_mapping = self.ordinal_encoder.category_mapping

        mappings_out = []
        for switch in ordinal_mapping:
            values = switch.get('mapping')
            col = switch.get('col')

            column_mapping = self.fit_helmert_coding(col, values, self.handle_missing, self.handle_unknown)
            mappings_out.append({'col': col, 'mapping': column_mapping, })

        self.mapping = mappings_out
        
        X_temp = self.transform(X, override_return_df=True)
        self.feature_names = X_temp.columns.tolist()
        
        

        if self.drop_invariant:
            self.drop_cols = []
            generated_cols = util.get_generated_cols(X, X_temp, self.cols)
            self.drop_cols = [x for x in generated_cols if X_temp[x].var() <= 10e-5]
            try:
                [self.feature_names.remove(x) for x in self.drop_cols]
            except KeyError as e:
                if self.verbose > 0:
                    print("Could not remove column from feature names."
                          "Not found in generated cols.\n{}".format(e))

        return self

    def transform(self, X, override_return_df=False):
        """Perform the transformation to new categorical data.

        Parameters
        ----------

        X : array-like, shape = [n_samples, n_features]

        Returns
        -------

        p : array, shape = [n_samples, n_numeric + N]
            Transformed values with encoding applied.

        """

        if self.handle_missing == 'error':
            if X[self.cols].isnull().any().any():
                raise ValueError('Columns to be encoded can not contain null')

        if self._dim is None:
            raise ValueError('Must train encoder before it can be used to transform data.')

        # first check the type
        X = util.convert_input(X)

        # then make sure that it is the right size
        if X.shape[1] != self._dim:
            raise ValueError('Unexpected input dimension %d, expected %d' % (X.shape[1], self._dim, ))

        if not list(self.cols):
            return X

        X = self.ordinal_encoder.transform(X)

        if self.handle_unknown == 'error':
            if X[self.cols].isin([-1]).any().any():
                raise ValueError('Columns to be encoded can not contain new values')

        X = self.helmert_coding(X, mapping=self.mapping)

        if self.drop_invariant:
            for col in self.drop_cols:
                X.drop(col, 1, inplace=True)

        if self.return_df or override_return_df:
            return X
        else:
            return X.values


    @staticmethod
    def fit_helmert_coding(col, values, handle_missing, handle_unknown):
        if handle_missing == 'value':
            values = values[values > 0]

        values_to_encode = values.values

        if len(values) < 2:
            return pd.DataFrame(index=values_to_encode)

        if handle_unknown == 'indicator':
            values_to_encode = np.append(values_to_encode, -1)

        helmert_contrast_matrix = Helmert().code_without_intercept(values_to_encode)
        df = pd.DataFrame(data=helmert_contrast_matrix.matrix, index=values_to_encode,
                          columns=[str(col) + '_%d' % (i,) for i in
                                   range(len(helmert_contrast_matrix.column_suffixes))])

        if handle_unknown == 'return_nan':
            df.loc[-1] = np.nan
        elif handle_unknown == 'value':
            df.loc[-1] = np.zeros(len(values_to_encode) - 1)

        if handle_missing == 'return_nan':
            df.loc[values.loc[np.nan]] = np.nan
        elif handle_missing == 'value':
            df.loc[-2] = np.zeros(len(values_to_encode) - 1)

        return df

    @staticmethod
    def helmert_coding(X_in, mapping):
        """
        """

        X = X_in.copy(deep=True)

        cols = X.columns.values.tolist()

        X['intercept'] = pd.Series([1] * X.shape[0], index=X.index)

        for switch in mapping:
            col = switch.get('col')
            mod = switch.get('mapping')

            base_df = mod.reindex(X[col])
            base_df.set_index(X.index, inplace=True)
            X = pd.concat([base_df, X], axis=1)

            old_column_index = cols.index(col)
            cols[old_column_index: old_column_index + 1] = mod.columns

        cols = ['intercept'] + cols

        return X.reindex(columns=cols)

    def get_feature_names(self):
        """
        Returns the names of all transformed / added columns.

        Returns
        -------
        feature_names: list
            A list with all feature names transformed or added.
            Note: potentially dropped features are not included!

        """

        if not isinstance(self.feature_names, list):
            raise ValueError('Must fit data first. Affected feature names are not known before.')
        else:
            return self.feature_names


# -*- coding: utf-8 -*-


class InputParamPath:
    """
    Define url path for different parameters
    """

    notebook = "/inputParameter/getInputParameterListForReferenceIdAndReferenceType/referenceId={notebook_id}/inputReferenceType=NOTEBOOK"

    project = "/inputParameter/getInputParameterListForProjectId/projectId={project_id}"

    user = "/inputParameter/getInputParameterListForUserName/"

    model = "/inputParameter/getInputParameterListForReferenceIdAndReferenceType/referenceId={version_id}/inputReferenceType=MODEL"

    input_params_path = "/inputParameter/getInputParameterValuesForReferenceIdAndInputReferenceType/referenceId={reference_id}/inputReferenceType={inputReferenceType}"


class InputParamReferenceType:

    ref_type_model = "MODEL"
    ref_type_notebook = "NOTEBOOK"
    ref_type_project = "PROJECT"
    ref_type_user = "USER"



import os
import time
import cloudpickle as pickle
import pandas as pd
from datetime import datetime, timedelta, timezone



class TimeTravelLoaderMixin:
    """
    Enables Retrieval of Objects across time range from serving models.
    """
    def __init__(self,
                 *args,
                 **kwargs):
        self.keys_to_exclude = ("score_func.pkl",)

    def load_from_time(self,
                       ml_model_id: str,
                       ml_version_id: str,
                       from_timestamp: datetime,
                       to_timestamp: datetime,
                       add_data_to_catalog: bool = False,
                       file_name: str = None,
                       model_path: str = None,
                       ):
        """
        This will retrieve payloads and valid payload from selected time frame.

        :param ml_model_id: Model Id String
        :param ml_version_id: Version Id String
        :param from_timestamp: From TimeStamp Must Be Timezone Aware (UTC) ex: datetime.now(timezone.utc)
        :param to_timestamp:  To TimeStamp Must Be Timezone Aware (UTC) ex: datetime.now(timezone.utc)
        :param add_data_to_catalog:
            if True data will be persisted in Mosaic AI DATA Section.
            Requires tabulate=True at Init.
            Note: This will first download data into your workspace.
        :param file_name:
            If Provided Will Upload Data With Same Name in Mosaic AI Data Catalog.
            Requires tabulate=True at Init.
        :param model_path: mount path for model data
        :return: Bool
        """
        self.model_dir_path = f"{model_path}/{ml_model_id}/{ml_version_id}/"
        self.all_keys = os.listdir(self.model_dir_path)
        self.only_pickled_objects = [objects_ for objects_ in self.all_keys if objects_.endswith("pkl")
                                     and not objects_.endswith(self.keys_to_exclude)]

        # loading files based on it's last modified time
        self.keys_to_load = [pickled_items for pickled_items in self.only_pickled_objects \
                             if from_timestamp <=  time.ctime(os.path.getmtime(os.path.join(self.model_dir_path, pickled_items))) <= to_timestamp]


        self.temp_score_collection_ = []

        for item in self.keys_to_load:
            self.temp_score_collection_ \
                .extend(pickle.load(open(os.path.join(self.model_dir_path, item), "rb")))

        self.add_response(score_response=self.temp_score_collection_)

        if add_data_to_catalog and self.tabulate_frame is not None:
            from connector.mosaicio import MosaicioConnector
            self.io_connector = MosaicioConnector()
            name = file_name or "{}_{}.csv".format(datetime.strptime(from_timestamp, '%Y_%m_%d'),
                                                    datetime.strptime(to_timestamp, '%Y_%m_%d'))
            self.tabulate_frame.to_csv(name, index=False)
            self.io_connector.upload(name)


        return True



# -*- coding: utf-8 -*-
import json
from .utils import NumpyArrayEncoder


def load_model(model_path):
    import spacy

    return spacy.load(model_path)


def dump_model(model, path):
    model.to_disk(path)


def get_model_structure(model_obj):
    ml_model_class = model_obj.__dict__
    ml_model_class["class"] = str(model_obj.__class__)[8:-2:]
    return json.dumps(ml_model_class, cls=NumpyArrayEncoder)



# -*- coding: utf-8 -*-
class Notebook:
    url = "http://notebooks-api:5000/notebooks/api"
    git_access_api = "/v1/git-repo/Enabled"
    get_repo_by_id_api = "/v1/git-repo-by-id/"

    
class MosaicVersionControl:
    url = "http://mosaic-version-control:9000/vcs/api"
    git_tag = "/v1/repo/{repo}/{version_id}/gittag?repo_id={repo_id}&branch={branch}"


class Headers:
    x_project_id = "X-Project-Id"
    x_username = "X-Auth-Username"
    x_userid = "X-Auth-Userid"
    x_email_id = "X-Auth-Email"



# -*- coding: utf-8 -*-
import os

import numpy as np
import pandas as pd

from .constants import Client, Headers, MLModelType
from ..build_time_metrics import metrics

from IPython.display import clear_output


def read_jwt():
    config_file = os.path.expanduser(Client.config_file)
    if os.path.exists(config_file):
        with open(config_file) as f:
            jwt = f.read().strip("\n")
        return jwt
    elif os.getenv("TOKEN"):
        jwt = os.getenv("TOKEN")
        return jwt
    else:
        print("Warning - JWT Token not set")


def get_headers():
    jwt = read_jwt()
    return {
        Headers.authorization: f"Token {jwt}",
        Headers.x_project_id: os.environ.get("PROJECT_ID"),
    }


def try_or(fn):
    try:
        out = fn()
        return out
    except:
        return None


def get_model_type_handler(model_type):
    if model_type == MLModelType.classification:
        return metrics.Classification
    if model_type == MLModelType.regression:
        return metrics.Regression


def get_array_type_handler(y_true, y_pred):
    if isinstance(y_true, list) and isinstance(y_pred, list):
        return "list"
    if isinstance(y_true, np.ndarray) and isinstance(y_pred, np.ndarray):
        return "ndarray"
    if isinstance(y_true, pd.core.series.Series) and isinstance(
        y_pred, pd.core.series.Series
    ):
        return "series"
    if isinstance(y_true, pd.core.frame.DataFrame) and isinstance(
        y_pred, pd.core.frame.DataFrame
    ):
        return "df"


def get_pred_and_true_values(y_true, y_pred):
    if len(y_true.columns) == 1 and len(y_pred.columns) == 1:
        return y_true.values, y_pred.values


def update_progress(progress):
    bar_length = 70
    if isinstance(progress, int):
        progress = float(progress)
    if not isinstance(progress, float):
        progress = 0
    if progress < 0:
        progress = 0
    if progress >= 1:
        progress = 1
    block = int(round(bar_length * progress))
    clear_output(wait=True)
    print("Calculating build time metrics\n")
    text = "Progress: {0} {1:.1f}%".format(
        "█" * block + " " * (bar_length - block), progress * 100
    )
    print(text)




import sys
import os
from importlib import import_module
from sklearn.feature_extraction import text 
from sklearn.base import TransformerMixin, BaseEstimator
from sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer, HashingVectorizer
from sklearn.preprocessing import OneHotEncoder
from importlib import import_module
from sklearn.compose import ColumnTransformer
import random
import re
import nltk
nltk.download('stopwords')
import pandas as pd
from bs4 import BeautifulSoup
from nltk.corpus import stopwords
from nltk.stem.snowball import SnowballStemmer
# init stemmer
stemmer = SnowballStemmer('english')
stops = list(stopwords.words('english'))
def preprocessor(text):
    # 1. Delete HTML 
    review_text = BeautifulSoup(text, 'html.parser').get_text()
    # 2. Make a space
    letters_only = re.sub('[^a-zA-Z]', ' ', review_text)
    # 3. lower letters
    words = letters_only.lower().split()
    # 5. Stopwords 
    meaningful_words = [w for w in words if not w in stops]
    # 6. Stemming
    stemming_words = [stemmer.stem(w) for w in meaningful_words]
    # 7. space join words
    return( ' '.join(stemming_words))



import_loader = {
    'CountVectorizer': CountVectorizer(preprocessor=preprocessor,stop_words=stops)
}



def load_scoring_function(scoring_func):
    """
    converts mymodule.myfunc in the myfunc
    object itself so tpot receives a scoring function
    """
    if scoring_func and ("." in scoring_func):
        try:
            module_name, func_name = scoring_func.rsplit('.', 1)

            module_path = os.getcwd()
            sys.path.insert(0, module_path)
            scoring_func = getattr(import_module(module_name), func_name)
            sys.path.pop(0)

            print('manual scoring function: {}'.format(scoring_func))
            print('taken from module: {}'.format(module_name))
        except Exception as e:
            print('failed importing custom scoring function, error: {}'.format(str(e)))
            raise ValueError(e)

    return scoring_func

class IdentityTransformer(TransformerMixin, BaseEstimator):
    """Identity-transformer for doing literally nothing"""
    def fit(self, X, y=None, **fit_params):
        return self
    def transform(self, X):
        return X

class PreprocessTransformer(TransformerMixin):
    
    def __init__(self, numeric_columns=[], categorical_columns=[], text_columns=[],
    text_transformer = 'CountVectorizer'):
        self.numeric_columns = numeric_columns
        #self.categorical_columns = categorical_columns
        self.text_columns = text_columns
        self.text_transformer = text_transformer
        #self.categorical_transformer = categorical_transformer
        
    def _setup_columns(self):
        def text_to_list(text):
            if type(text) is list:
                return text
            return [text]
        self.numeric_columns = text_to_list(self.numeric_columns)
        #self.categorical_columns = text_to_list(self.categorical_columns)
        self.text_columns = text_to_list(self.text_columns)

        column_list = []
        if len(self.text_columns) > 0:
            load_func = import_loader.get(self.text_transformer, self.text_transformer)
            #load_func = load_func.__class__
            if isinstance(load_func, str):
                load_func = load_scoring_function(load_func)
            for idx, text in enumerate(self.text_columns):
                column_list.append(('text' + str(idx), load_func , text))
        
        if len(self.numeric_columns) > 0:
            column_list.append(('numeric', IdentityTransformer(), self.numeric_columns))
        '''
        if len(self.categorical_columns) > 0:
            load_func = import_loader.get(self.categorical_transformer, self.categorical_transformer)
            if isinstance(load_func, str):
                load_func = load_scoring_function(load_func)
            column_list.append(('categorical', load_func, self.categorical_columns))
        '''    
        self.column_transformer = ColumnTransformer(column_list)

    def fit(self, X, y=None):
        self._setup_columns()
        self.column_transformer.fit(X, y)
        return self
    def transform(self, X):
        return self.column_transformer.transform(X)  



# -*- coding: utf-8 -*-
import functools
import sys
from flask import current_app
import logging


def log_decorator(_func=None):
    """Logging Decorator"""

    def log_decorator_info(func):
        @functools.wraps(func)
        def log_decorator_wrapper(*args, **kwargs):
            """log function begining"""

            try:
                LOGGER = current_app.logger
            except:
                LOGGER = logging.getLogger("default_logger")

            LOGGER.info("Begin function: {0}".format(func.__name__))
            try:
                value = func(*args, **kwargs)
                LOGGER.info("End function: {0}".format(func.__name__))
            except:
                LOGGER.error(f"Exception: {str(sys.exc_info()[1])}")
                LOGGER.info("End function: {0} with exception".format(func.__name__))
                raise
            return value

        return log_decorator_wrapper

    if _func is None:
        return log_decorator_info
    return log_decorator_info(_func)



# -*- coding: utf-8 -*-
from uuid import uuid4
import requests
from .constants import *
from ..headers.utils import generate_headers
import logging


def uuid_generator():
    """
    Method to generate uuid
    :return: uuid
    :rtype: str
    """
    _uuid = uuid4()
    return str(_uuid)



def notification_alert(notification_type, message_status, notification_url, headers):

    """
    Calls lens-manage-console api for storing and triggering notifications

    Args:
            notification_type (string): Notebooks/AutoML/Model/Schedule
            headers (dict): headers to be passed to API
            message_status (string): Description of the notification
            notification_url (string): redirect url

    Returns:
            response after hitting the api
    """


    notification_payload = {
        "user_id": headers["X-Auth-Username"],
        "status": "Success",
        "uuid": uuid_generator(),
        "notificationURL": notification_url,
        "notificationDescription": message_status,
        "notificationType": notification_type,  # the real deal
        "createdBy": headers["X-Auth-Username"],
        "notificationDetails": {
            "senderName": headers["X-Auth-Username"],
            "accessType": "FullAccess"
        }
    }

    final_headers = {
        "X-Auth-Username": headers["X-Auth-Username"],
        "X-Project-Id": headers["X-Project-Id"],
        "X-Auth-Userid": headers["X-Auth-Userid"],
        "X-Auth-Email": headers["X-Auth-Email"]
    }

    response = requests.post(Url.notification_url, json=notification_payload, headers=final_headers)
    return response


def fetch_line_manager_details(userid, email, username, project_id):
    line_manager_email = ""
    try:
        logging.info("The username of the user is : %s", userid)
        line_manager_url = Url.LINE_MANAGER_URL + userid
        line_manager_response = requests.get(line_manager_url, headers=generate_headers(userid, email, username, project_id))
        if line_manager_response.status_code == 200:
            line_manager_response_json = line_manager_response.json()
            logging.info(line_manager_response_json)
            line_manager_email = line_manager_response_json["lineManagerDetails"]["emailId"]
            logging.info("Line manager email id : %s", line_manager_email)
        elif line_manager_response.status_code == 404:
            logging.info("Line Manager Details are unavailable")
        else:
            logging.error("Unknown Error while fetching Line-Manager Info : %s", line_manager_response.text)

        return line_manager_email

    except Exception as ex:
        logging.exception(ex)
        return line_manager_email



import datetime
import json
import pandas as pd
from uuid import uuid4
from collections.abc import Iterable, Iterator
from typing import Union, List
from .response import ScoreResponse
from .constants import ScoreResponseCollectionLabel as SCRLabel
from .constants import ScoreResponseLabel as SCLabel
from .mixins import TimeTravelLoaderMixin
from IPython.display import display


class OrderIterator(Iterator):
    """
    Concrete Iterators implement various traversal algorithms. These classes
    store the current traversal position at all times.
    """

    """
    `_position` attribute stores the current traversal position. An iterator may
    have a lot of other fields for storing iteration state, especially when it
    is supposed to work with a particular kind of collection.
    """
    purpose = "<<OrderIterator>>"
    _position: int = None

    def __repr__(self):
        return self.purpose

    """
    This attribute indicates the traversal direction.
    """
    _reverse: bool = False

    def __init__(self, collection, reverse: bool = False) -> None:
        self._collection = collection
        self._reverse = reverse
        self._position = -1 if reverse else 0

    def __next__(self):
        """
        The __next__() method must return the next item in the sequence. On
        reaching the end, and in subsequent calls, it must raise StopIteration.
        """
        try:
            value = self._collection[self._position]
            self._position += -1 if self._reverse else 1
        except IndexError:
            raise StopIteration()

        return value


class ScoreResponseCollection(TimeTravelLoaderMixin,
                              Iterable):
    """
    Concrete Collections provide one or several methods for retrieving fresh
    iterator instances, compatible with the collection class.
    """
    purpose = "<<ScoreResponseCollection>>"

    def __repr__(self):
        return self.purpose

    def __init__(self,
                 collection: List[ScoreResponse] = [],
                 mode="order",
                 tabulate=True,
                 tabulate_serialize=False,
                 score_cls=None,
                 *args,
                 **kwargs
                 ) -> None:
        super(ScoreResponseCollection, self).__init__(*args, **kwargs)

        self.tabulate = tabulate
        self.tabulate_serialize = tabulate_serialize
        self.score_cls = score_cls

        if tabulate:
            self.tabulate_frame = pd.DataFrame(columns=SCRLabel.SCORE_RESPONSE_FIELDS,
                                               )
            self.tabulate_frame["timestamp"] = pd.to_datetime(self.tabulate_frame["timestamp"],
                                                              format=SCLabel.TIMESTAMP_FORMAT)

        self._collection = []
        self.add_response(score_response=collection)
        self._mode = mode.lower()
        self.modes_available = {"order": OrderIterator,
                                }
        self._iterator_construct = self.modes_available.get(self._mode, OrderIterator)

        self.rp_payload_df: DataFrame = None
        self.clean_payload_df: DataFrame = None

    def __iter__(self) -> OrderIterator:
        """
        The __iter__() method returns the iterator object itself, by default we
        return the iterator in ascending order.
        """
        return self._iterator_construct(self._collection)

    def get_reverse_iterator(self) -> OrderIterator:
        return self._iterator_construct(self._collection, True)

    def add_response(self,
                     score_response: Union[ScoreResponse, List[ScoreResponse]]):

        if isinstance(score_response, ScoreResponse):
            score_response = [score_response]

        if all(isinstance(x, ScoreResponse) for x in score_response):
            self._collection.extend(score_response)
            if self.tabulate and self.tabulate_serialize:
                from .encoder import ScoreResponseEncoder

                serialize_ = json.dumps(score_response, cls=ScoreResponseEncoder)
                self.tabulate_frame = self.tabulate_frame.append(json.loads(serialize_),
                                                                 ignore_index=True)
            elif self.tabulate:
                for each_score in score_response:
                    self.tabulate_frame = self.tabulate_frame.append(each_score.to_dict(),
                                                                     ignore_index=True)

    def extend(self, collection):
        """
        :param collection: Must Be Instance of ScoreResponseCollection Iterable
        :return:
        """
        if type(collection) is type(self):
            self.add_response(score_response=list(collection))
        return

    def extract_rp_payload(self):
        if self.tabulate and self.score_cls:
            _valid_rp_payloads = self.tabulate_frame.loc[self.tabulate_frame['rp_payload_validated'] == True]['request_processed_payload']
            cols_ = list(self.score_cls.each_request_processed_schema.keys())
            self.rp_payload_df = pd.DataFrame([pd.Series(x, index=cols_) for x in _valid_rp_payloads],
                                              columns=cols_)

            display(self.rp_payload_df)
        else:
            print("Flag Must Be Set at Init: ['tabulate', 'score_cls'] ")

    def extract_clean_payload(self):
        if self.tabulate and self.score_cls:
            _valid_clean_payloads = self.tabulate_frame.loc[self.tabulate_frame['clean_payload_validated'] == True]['clean_payload']
            cols_ = list(self.score_cls.each_pre_processed_schema.keys())
            self.clean_payload_df = pd.DataFrame([pd.Series(x, index=cols_) for x in _valid_clean_payloads],
                                              columns=cols_)

            display(self.clean_payload_df)
        else:
            print("Flag Must Be Set at Init: ['tabulate', 'score_cls'] ")



# -*- coding: utf-8 -*-

"""Combine dataframes."""


class CombineDFs(object):
    """Combine two DataFrames."""

    @property
    def __name__(self):
        """Instance name is the same as the class name."""
        return self.__class__.__name__



class Url:
    # lens_manage_console_url = "http://lens-manageconsole:80/lens-manageconsole/insight-hub/notification-management/notification-add"
    notification_url = "http://notification-service/notification/v1/notification-add"
    WHOAMI_URL = "http://user-management:8094/usermgm/user/whomi"
    LINE_MANAGER_URL = "http://user-management:8094/usermgm/accessRequest/getManagerDetailsByUsersBankId/bankId="



from .constants import ScoreLabel
from collections import OrderedDict
import numpy as np


class GenericValidatorMixin:
    def update_validation_box(self, filter_condition, update_column, value):
        self.validation_box.loc[
            self.validation_box["VALIDATION"] == filter_condition,
            update_column
        ] = value
        return True


class PayloadSchemaValidatorMixin(GenericValidatorMixin):
    def validate_with_schema_payload(self,
                                     schema_variable: OrderedDict,
                                     validate_with_schema=None):

        if isinstance(schema_variable, OrderedDict):
            if hasattr(validate_with_schema, 'tolist') and callable(validate_with_schema.tolist):
                # Either Torch Tensor or Numpy
                validate_with_schema = validate_with_schema.tolist()
            elif hasattr(validate_with_schema, 'numpy') and callable(validate_with_schema.numpy):
                # Either Tensorflow Tensor
                validate_with_schema = validate_with_schema.numpy().tolist()

            val_ = all(map(lambda obj_, dtype: isinstance(obj_, dtype),
                           list(validate_with_schema),
                           list(schema_variable.values())
                           ))

            if not val_:
                print("Invalid: {} Expected {}".format(validate_with_schema,
                                                       list(schema_variable.values())))
                return False
        else:
            return False
        return True


class RequestValidatorMixin(GenericValidatorMixin):
    def validate_request(self, request_processed_input):
        if isinstance(request_processed_input, tuple):
            self.update_validation_box(ScoreLabel.Request_Return_Type, "PASSED", True)
        else:
            return
        self.update_validation_box(ScoreLabel.Request_Return_Type, "SKIPPED", False)

        if len(request_processed_input) == 2:
            self.update_validation_box(ScoreLabel.Request_Return_Shape, "PASSED", True)
            self.update_validation_box(ScoreLabel.Request_Return_Shape, "SKIPPED", False)

            n_input, payload = request_processed_input
        else:
            return

        if n_input > 1 and not isinstance(payload, list):
            self.update_validation_box(ScoreLabel.Request_Many_Payload_DType, "PASSED", False)
        self.update_validation_box(ScoreLabel.Request_Many_Payload_DType, "SKIPPED", False)




#! -*- coding: utf-8 -*-
# -*- coding: utf-8 -*-
""" This module contains different exceptions used by the mosaic AI serving """


class MosaicException(Exception):
    """ Common/Custom exceptions """

    code = 500
    message = "Internal Server Error"

    def __init__(self, message, code):
        self.message = message
        self.code = code



# -*- coding: utf-8 -*-
"""
This Package Defines core structure for Revamped Model Scoring.
===================================================================

Model scoring version 2.0 Has replaced old linear scoring function.
The revamped scoring follows three steps:
    1. Request Level Segregation where user seggregates bulk payload.
    2. Preprocessing Level where data preprocessing takes place.
    3. Prediction Level Where Inference happens on using single sample at a time.

Returns: <<ScoreResponse>>

base.py: Core Logic For Defining Scoring Abstract Class <<ScoreBase>>
response.py: Defines Response Schema <<ScoreResponse>>

encoder.py:
    - ScoreResponseEncoder : [<<ScoreResponse>>,..] object to JSON Convertor
    - ScoreResponseDecoder: JSON -> To [<<ScoreResponse>>, .. ] Convertor
iterators.py: <<ScoreResponseCollection>> holds all the score response and makes it iterable
mixins.py: <<TimeTravelLoaderMixin>>

constants.py: Constants
utils.py: Utility
validators.py:  Validations

"""

__author__ = "Shivam Chaurasia (10670364)"
__description__ = "Revamped Model Scoring Package"
__version__ = "2.0"
__deprecated__ = False
__date__ = "26 Feb, 2021"



# -*- coding: utf-8 -*-

"""Handler for PySpark Model Flavor."""

import os
import json
from .utils import NumpyArrayEncoder

# from pyspark.sql import SparkSession

# spark = SparkSession.builder.appName(__name__).getOrCreate()


class ClassCreator(object):
    """Create class instance from its string representation."""

    @classmethod
    def get_class(cls, class_name):
        """
        Get an instance of class from class_name.

        :param class_name:
        :return:
        """
        parts = class_name.split(".")
        module = ".".join(parts[:-1])
        import_module = __import__(module)
        for comp in parts[1:]:
            import_module = getattr(import_module, comp)
        return import_module


def get_model_type(model_path):
    """Get type of pyspark model."""
    metadata_path = os.path.join(model_path, "metadata")
    metadata_files = os.listdir(metadata_path)
    metadata_file = [x for x in metadata_files if x.startswith("part-000")]
    if not metadata_file:
        raise FileNotFoundError("Model metadata file not found")
    metadata_file = metadata_file[0]
    with open(os.path.join(metadata_path, metadata_file), "r") as f_handle:
        model_metadata = f_handle.read()
    model_metadata = json.loads(model_metadata)
    model_class = model_metadata.get("class")
    if not model_class:
        raise AttributeError('Attribute "class" not found in model metadata')
    model_class = model_class.replace("org.apache.spark", "pyspark")
    return model_class


def load_model(model_path):
    """Load model from disk."""
    model_type = get_model_type(model_path)
    model_handler = ClassCreator.get_class(model_type)
    model_instance = model_handler.load(model_path)
    return model_instance


def dump_model(model, path):
    """Save the model to disk."""
    model.save(path)
    import time
    time.sleep(2)


def get_model_structure(model_obj):
    ml_model_class = model_obj.__dict__
    ml_model_class["class"] = str(model_obj.__class__)[8:-2:]
    return json.dumps(ml_model_class, cls=NumpyArrayEncoder)



# -*- coding: utf-8 -*-
import json


def load_model(frozen_graph_path):
    """
    Load the model from frozen inference graph.

    Args:
        frozen_graph_path   :   path to frozen graph (.pb)

    Return:
        frozen graph that can be used directly for inference
    """
    import tensorflow as tf
    if tf.__version__[:1] == "2":
        detection_graph = tf.keras.models.load_model(frozen_graph_path)
    else:
        detection_graph = tf.Graph()
        with detection_graph.as_default():
            od_graph_def = tf.GraphDef()
            with tf.gfile.GFile(frozen_graph_path, "rb") as fid:
                serialized_graph = fid.read()
                od_graph_def.ParseFromString(serialized_graph)
                tf.import_graph_def(od_graph_def, name="")

    return detection_graph


def dump_model(frozen_graph, path):
    """
    Save Tensorflow Frozen Inference Graph.

    Args:
        frozen_graph    :   tensorflow frozed inference graph
        path            :   path to save the frozen inference graph (.pb)
    """
    import tensorflow as tf
    if tf.__version__[:1] == "2":
        tf.keras.models.save_model(frozen_graph, path)
    else:
        with tf.gfile.GFile(path, "wb") as f:
            f.write(frozen_graph.as_graph_def().SerializeToString())


def get_model_structure(model_obj):
    ml_model_class = {"class": str(model_obj.__class__)[8:-2:]}
    return json.dumps(ml_model_class)



# -*- coding: utf-8 -*-
from kubernetes.client import CustomObjectsApi, CoreV1Api
from kubernetes import config
import os
import shutil
import math
from .utils import fetching_cpu_value
import requests
from .constants import GenericConstants
from typing import Dict


if shutil.which("minikube"):
    config.load_kube_config()
elif os.getenv("KUBERNETES_SERVICE_HOST"):
    config.load_incluster_config()
else:
    pass

custom_object = CustomObjectsApi()
core_v1 = CoreV1Api()


def fetch_resource_limitscaling_guarantee(cpu, memory, resource_extra, cpulimitpercentage, memorylimitpercentage):

    """limit vertical scaling"""
    limit_cpu_specified = cpu
    if limit_cpu_specified[-1:] == "m":
        limit_cpu = int(limit_cpu_specified[:-1])
    else:
        limit_cpu = int(limit_cpu_specified)
        limit_cpu = limit_cpu * 1000
    cpulimitpercentage = int(cpulimitpercentage)
    limitcpu_scale = (limit_cpu * (100 + cpulimitpercentage)) / 100
    limitcpu_scale = str(math.trunc(limitcpu_scale))
    limitcpu_scale = limitcpu_scale + "m"
    limit_memory_specified = memory
    limit_memory = limit_memory_specified[:-2]
    if limit_memory_specified[-2:] == "Gi":
        limit_memory = int(limit_memory)
        limit_memory = limit_memory * 1000
    elif limit_memory_specified[-2:] == "Mi":
        limit_memory = int(limit_memory)
    limit_unit = "Mi"
    memorylimitpercentage = int(memorylimitpercentage)
    limitmemory_scale = str(math.trunc((limit_memory * (100 + memorylimitpercentage)) / 100))
    limitmemory_scale = limitmemory_scale + limit_unit
    if resource_extra == "nvidia":
        return {"nvidia.com/gpu": cpu, "memory": limitmemory_scale}
    if resource_extra == "amd":
        return {"amd.com/gpu": cpu, "memory": limitmemory_scale}
    return {"cpu": limitcpu_scale, "memory": limitmemory_scale}


def fetch_pod_metrics(pod_name, namespace, user_name):
    pod_metrics = {}

    resp = custom_object.get_namespaced_custom_object(
        "metrics.k8s.io", "v1beta1", namespace, "pods", pod_name
    )
    status = core_v1.read_namespaced_pod_status(namespace=namespace, name=pod_name)
    pod_status = status.status.phase
    pod_start_time = status.status.start_time

    for container in resp["containers"]:
        current_cpu_usage = container["usage"]["cpu"]
        current_memory_usage = container["usage"]["memory"]

        for container_resource in status.spec.containers:
            if container_resource.name == container["name"]:
                pod_container_limits = container_resource.resources.limits
                pod_container_requests = container_resource.resources.requests

                # Calculating CPU percentages

                # fetching cpu value
                cpu = fetching_cpu_value(current_cpu_usage)

                if "m" in pod_container_limits["cpu"]:
                    revised_pod_limits = int(
                        pod_container_limits["cpu"].replace("m", "")
                    )
                    core_limit = revised_pod_limits / 1000
                else:
                    core_limit = int(pod_container_limits["cpu"])

                core_usage_percentage = (cpu * 100) / core_limit

                # Converting memory usage from KB's to MB's
                memory_usage = current_memory_usage.replace("Ki", "")
                memory_usage_mb = int(memory_usage) / 1024

                data = {
                    container_resource.name: {
                        "cpu_cores_used": str(cpu),
                        "memory": str(memory_usage_mb),
                        "pod_status": pod_status,
                        "limits": pod_container_limits,
                        "requests": pod_container_requests,
                        "cpu_percent_used": str(core_usage_percentage),
                        "created_by": user_name,
                        "pod_start_time": pod_start_time,
                    }
                }
                pod_metrics.update(data)

    return pod_metrics


def volume_count(project_id, username,
                 project_manager_base_url = "http://mosaic-console-backend/mosaic-console-backend"):
    url = f"{project_manager_base_url}/secured/api/pvc/project/{project_id}"
    url2 = f"{project_manager_base_url}/secured/api/pvc/project/all"
    volume = []
    headers = {"X-Auth-Username": username}
    for item in [url, url2]:
        response = requests.get(item, headers=headers)
        response = response.json()
        if response:
            for x in range(len(response)):
                desc = {
                    "name": response[x]["pvcName"],
                    "persistentVolumeClaim": {"claimName": response[x]["pvcName"]},
                }
                volume.append(desc)
    return volume

def volume_custom(project_id, pvc_name):
    volume = []
    desc = {
        "name": project_id,
        "persistentVolumeClaim": {"claimName": pvc_name},
    }
    volume.append(desc)
    return volume

def volume_custom_mount(project_id=None, minio_bucket=None, resource_quota_full=None):
    volume = []
    desc = {
        "name": project_id,
        "mountPath": "/data",
        "subPath": f"{minio_bucket}/{project_id}/{project_id}-Data"
    }
    if resource_quota_full:
        desc["readOnly"]=True
    volume.append(desc)
    return volume

def volume_mount_count(project_id, username,
                       project_manager_base_url = "http://mosaic-console-backend/mosaic-console-backend"):
    """
    :param project_id: Project Id
    :param username: Username
    :param project_manager_base_url: Base url for project manager service
    :return: It returns list of volumes(pvcname, mountpath) configured for given project
    """
    url = f"{project_manager_base_url}/secured/api/pvc/project/{project_id}"
    url2 = f"{project_manager_base_url}/secured/api/pvc/project/all"
    volume = []
    headers = {"X-Auth-Username": username}
    for item in [url, url2]:
        response = requests.get(item, headers=headers)
        response = response.json()
        if response:
            for x in range(len(response)):
                desc = {
                    "name": response[x]["pvcName"],
                    "mountPath": response[x]["mountpath"],
                }
                volume.append(desc)
    return volume


def attach_default_volume(project_id=None, shared_volumes=None, storage_class=None):
    """
    This method is used to attach default volume mount to the spawned containers
    :return:
    """
    return {
        "type": storage_class,
        "path": shared_volumes
        + GenericConstants.project_data_dir.format(project_id, project_id),
        "name": project_id,
    }


def attach_snapshot_volume(
    project_id=None, shared_volumes=None, storage_class=None, text=None, snap=None
):
    """
    This method is used to attach default volume mount to the spawned containers
    :return:
    """
    if snap == GenericConstants.default:
        return {
            "type": storage_class,
            "path": shared_volumes
            + GenericConstants.all_snapshot_data.format(project_id, project_id),
            "name": text,
        }
    return {
        "type": storage_class,
        "path": shared_volumes
        + GenericConstants.snapshot_data_fir.format(project_id, project_id, snap),
        "name": text,
    }


def attach_default_volume_mount(project_id=None, resource_quota_full=None):
    """
    This method is used to attach default volume mount to the spawned containers
    :return:
    """
    if resource_quota_full:
        return {"name": project_id, "mountPath": "/data", "readOnly": True}
    return {"name": project_id, "mountPath": "/data"}


def attach_snapshot_volume_mount(project_id=None, snapshot_type=None, snapshot_id=None, minio_bucket=None, resource_quota_full=None):
    """
    This method is used to attach default volume mount to the spawned containers
    :return:
    """
    if snapshot_type == GenericConstants.default:
        return {"name": project_id, "mountPath": "/input", "subPath": f"{minio_bucket}/{project_id}/{project_id}-Snapshot/{snapshot_id}", "readOnly": True}
    if resource_quota_full:
        return {"name": project_id, "mountPath": "/output", "subPath": f"{minio_bucket}/{project_id}/{project_id}-Snapshot/{snapshot_id}","readOnly": True}
    return {"name": project_id, "mountPath": "/output", "subPath": f"{minio_bucket}/{project_id}/{project_id}-Snapshot/{snapshot_id}"}


def fetch_resource_request_limit(
    requested_cpu: str,
    requested_memory: str,
    cpu_resource_percent: int,
    memory_resource_percent: int,
    resource_extra: str = None,
) -> Dict:
    """
    Limits resource request to resource_percent of the requested resource
    :param requested_cpu: Value of requested cpu by user. ex: cpu: "100m", "1"
    :param requested_memory:  Value of requested memory by user. ex: cpu: "256Mi", "2Gi"
    :param resource_percent: percent value from config-map
    :param resource_extra: GPU if present, else None
    :return: dict of request cpu and memory
    """

    gpu = requested_cpu
    units = {"m": "m", "Gi": "Gi", "Mi": "Mi"}
    cpu_resource_percentage = int(cpu_resource_percent)
    memory_resource_percentage = int(memory_resource_percent)
    request_memory = int(requested_memory[:-2])

    if requested_cpu[-1:] == units["m"]:
        requested_cpu = int(requested_cpu[:-1])
    else:
        requested_cpu = int(requested_cpu) * 1000
    requested_cpu = str(math.trunc((requested_cpu * cpu_resource_percentage) / 100))
    requested_cpu += units["m"]

    if requested_memory[-2:] == units["Gi"]:
        request_memory = request_memory * 1000
    requested_memory = str(math.trunc((request_memory * memory_resource_percentage) / 100))
    requested_memory += units["Mi"]

    resources_mapping = {
        "nvidia": {"nvidia.com/gpu": gpu, "memory": requested_memory},
        "amd": {"amd.com/gpu": gpu, "memory": requested_memory},
    }
    if resource_extra in resources_mapping.keys():
        return resources_mapping[resource_extra]

    return {"cpu": requested_cpu, "memory": requested_memory}



# -*- coding: utf-8 -*-
"""validators.py: Handler for handling all validations."""

__author__ = "Shivam Chaurasia (10670364)"
__description__ = "Implements ValidationHandler which composites all the validations present in validations.py"
__version__ = "0.1"

from .validations import *
from typing import List, Dict
import logging


class ValidationHandler(object):
    purpose = "Handler/MosaicAiClient/ValidatorEngine"

    def __repr__(self):
        return self.purpose

    def __init__(self,
                 mandatory_fields: Dict = None,
                 alphanum_uscore_excl: Dict = None,
                 if_present_validate_type: Dict = None,
                 if_present_sub_field_must_exist: Dict = None
                 ):
        '''
        Specify all the Parent and Subfield Name and Values for validations needs to be handled.
        ===============================================

        :param mandatory_fields
            : Dict[Name: Value]
            ex: {"variable_name": value, "variable_name": value, ..}
        :param alphanum_uscore_excl
            : Dict[Name: Value]
            ex: {"variable_name": value, "variable_name": value, ..}
        :param if_present_validate_type
            : Dict[Name:Tuple(DTypes)]
            ex:
            {
                "parent_var_name": tuple(parent_var_value, tuple(d_types,...)),
                "parent_var_name": tuple(parent_var_value, tuple(d_types,...)),
                ...
            }
        :param if_present_sub_field_must_exist
            : Dict[ParentName:Tuple(ParentValue, Dict[SubField: SubFieldValue])]
            ex:
            {
                "parent_var_name": (parent_var_value,
                                    {"sub_field_name": sub_field_value,
                                     "sub_field_name": sub_field_value,
                                     ...
                                     },..),
                "parent_var_name": (parent_var_value,
                                    {"sub_field_name": sub_field_value,
                                     "sub_field_name": sub_field_value,
                                     ...
                                     },..),
                ...
            }

        '''

        self._validations = dict()

        if mandatory_fields:
            self._validations.update({"mandatory_fields":
                                          MandatoryFields(mandatory_fields)})

        if alphanum_uscore_excl:
            self._validations.update({"alphanum_uscore_excl":
                                          ANumericUScoresExclusive(alphanum_uscore_excl)})
        if if_present_validate_type:
            self._validations.update({'if_present_validate_type':
                                          IfPresentTypeCheck(fields=if_present_validate_type)})

        if if_present_sub_field_must_exist:
            self._validations.update({'if_present_sub_field_must_exist':
                                          IfPresentSubfieldMustExist(fields=if_present_sub_field_must_exist)})

    def validate(self,
                 mandatory_fields=True,
                 alphanum_uscore_excl=True,
                 if_present_validate_type=True,
                 if_present_sub_field_must_exist=True):
        to_check = list()

        if mandatory_fields and "mandatory_fields" in self._validations.keys():
            to_check.append(self._validations["mandatory_fields"])

        if alphanum_uscore_excl and "alphanum_uscore_excl" in self._validations.keys():
            to_check.append(self._validations["alphanum_uscore_excl"])

        if if_present_validate_type and "if_present_validate_type" in self._validations.keys():
            to_check.append(self._validations["if_present_validate_type"])

        if if_present_sub_field_must_exist and "if_present_sub_field_must_exist" in self._validations.keys():
            to_check.append(self._validations["if_present_sub_field_must_exist"])

        for each_validation_test in to_check:
            status, message = each_validation_test.check()
            if status is False:
                return status, message

        return True, "validation_completed"



import os
from uuid import uuid4


def fetching_cpu_value(current_cpu_usage):
    if "n" in current_cpu_usage:
        cpu = int(current_cpu_usage.replace("n", "")) / 1000000000
    elif "m" in current_cpu_usage:
        cpu = int(current_cpu_usage.replace("m", "")) / 1000
    else:
        cpu = int(current_cpu_usage)
    return cpu


def create_job_name(file_path: str, instance_id: int = None) -> str:
    """
    Returns k8s compliant job name from filepath string
    :param file_path:
    :return: jobname str
    :param instance_id: in case instance_id is present add that in name instead of uuid
    :type instance_id: int
    """
    # get file name from file_path
    file_name = os.path.basename(file_path)
    # remove extensions from file name
    revised_notebook_name = os.path.splitext(file_name)[0]

    job_name = "{}-{}-job".format(revised_notebook_name.lower(), instance_id if instance_id else str(uuid4()))
    job_name = "".join(e for e in job_name if e == "-" or e.isalnum())
    if len(job_name) > 55:
        job_name = job_name[-55:]

    # first character of job name should always be alphanumeric
    if not job_name[0].isalnum():
        job_name = "0" + job_name[1:]
    return job_name



import json
import copy
from json import JSONEncoder, JSONDecoder
from datetime import datetime
from numpy import ndarray
from .response import ScoreResponse
from .constants import ScoreResponseLabel as ScoreRLBL


class ScoreResponseDecoder(JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, dct):
        _uuid = dct["_uuid"]
        _timestamp = datetime.strptime(dct["timestamp"],
                                       ScoreRLBL.TIMESTAMP_FORMAT)
        _request_processed_payload = dct["request_processed_payload"]
        _rp_payload_validated = dct["rp_payload_validated"]
        _clean_payload = dct["clean_payload"]
        _clean_payload_validated = dct["clean_payload_validated"]
        _score_response = dct["score_response"]
        _score_request = dct["score_request"]

        return ScoreResponse(uuid=_uuid,
                             timestamp=_timestamp,
                             score_request=_score_request,
                             request_processed_payload=_request_processed_payload,
                             rp_payload_validated=_rp_payload_validated,
                             clean_payload=_clean_payload,
                             clean_payload_validated=_clean_payload_validated,
                             score_response=_score_response)


class ScoreResponseEncoder(JSONEncoder):
    def default(self, o):
        self._object = copy.deepcopy(o.__dict__)

        self._object["timestamp"] = self.encode_timestamp()
        self._object["request_processed_payload"] = self.encode_request_processed_payload()
        self._object["rp_payload_validated"] = self.encode_rp_payload_validated()
        self._object["clean_payload"] = self.encode_clean_payload()
        self._object["clean_payload_validated"] = self.encode_clean_payload_validated()
        self._object["score_request"] = self.encode_score_request()

        return self._object

    def encode_clean_payload_validated(self):
        return bool(self._object["clean_payload_validated"])

    def encode_rp_payload_validated(self):
        return bool(self._object["rp_payload_validated"])

    def encode_timestamp(self):
        return self.check_value_dtype(self._object["timestamp"])

    def encode_request_processed_payload(self):
        return self.check_value_dtype(self._object["request_processed_payload"])

    def encode_clean_payload(self):
        return self.check_value_dtype(self._object["clean_payload"])

    def encode_score_request(self):
        return self.check_value_dtype(self._object["score_request"])

    def check_value_dtype(self, value):
        if isinstance(value, datetime):
            return value.strftime(ScoreRLBL.TIMESTAMP_FORMAT)
        elif isinstance(value, ndarray):
            return value.tolist()
        elif hasattr(value, 'tolist'):
            # Either Torch Tensor
            return value.tolist()
        elif hasattr(value, 'numpy') and callable(value.numpy):
            # Either Tensorflow Tensor
            return value.numpy().tolist()
        elif isinstance(value, (list, str)):
            return value
        return []




# -*- coding: utf-8 -*-
import json


def load_model(model_path):
    from tensorflow import keras

    return keras.models.load_model(model_path)


def dump_model(model, path):
    model.save(path)


def get_model_structure(model_obj):
    ml_model_class = json.loads(model_obj.to_json())
    ml_model_class["class"] = str(model_obj.__class__)[8:-2:]
    return json.dumps(ml_model_class)



# -*- coding: utf-8 -*-
class GenericConstants:
    shared_volumes = "shared_volume"
    slash = "/"
    project_data_dir = "{0}/{1}-Data"
    snapshot_data_fir= "{0}/{1}-Snapshot/{2}"
    default = "input"
    all_snapshot_data= "{0}/{1}-Snapshot"



# -*- coding: utf-8 -*-
from setuptools import find_packages, setup

extras_require = {
    "flavours": [
        "keras==2.2.4",
        "pyspark==2.4.6",
        "sklearn==0.0",
        "tensorflow==2.4.1",
        "tensorflow-probability==0.11.1",
        "torch==1.2.0",
        "torchvision==0.4.0",
        # changing the spacy version as 2.2.1 giving segmentation faults
        "spacy==2.2.2",
        # installing the scikit-learn version 0.21.3 since 0.22 is not stable
        "scikit-learn==0.21.3",
    ],
    "common": [
        "cloudpickle==1.2.2",
        "minio==4.0.18",
        "GitPython==3.1.30",
        "gitdb2==2.0.6",
        "ipython>=7.11.1",
    ],
    "empty": [
        "ipython>=7.11.1",
    ],

    "nb-template-serving": [
        "cloudpickle==1.2.2",
        # "minio==4.0.18",
        # "GitPython==3.0.4",
        # "gitdb2==2.0.6",
        # "ipython>=7.11.1",
        "PyYAML",
        # installing the scikit-learn version 0.21.3 since 0.22 is not stable
        ## Vulnerability fixes
        # "scikit-learn==0.24.0",
    ],
    "metrics": [
        "numpy>=1.16.5",
        "requests==2.22.0",
        "pandas==1.0.4",
        "sklearn==0.0",
        "PyYAML",
        # installing the scikit-learn version 0.21.3 since 0.22 is not stable
        "scikit-learn==0.21.3",
        "ipython>=7.11.1",
    ],
    "k8": [
        "kubernetes==9.0.0",
        "numpy>=1.16.5",
        "pandas==1.0.4",
        "sklearn==0.0",
        "scikit-learn==0.21.3",
    ],
    "mosaicautoml": [
        "numpy>=1.16.5",
        "pandas==1.0.4",
        "sklearn==0.0",
        "category-encoders==2.2.2",
        "nltk==3.5",
        "bs4==0.0.1",
    ],
    "r-jupyter": [
        "torch==1.8.0",
        "torchvision==0.8.0",
        "PyYAML",
        "scikit-learn==0.21.3"
    ],
    "most_common": [
        "ipython>=7.11.1",
        "alembic==1.0.11",
        "celery==5.2.2",
        "flasgger==0.9.4",
        "flask==2.2.5",
        "flask-migrate==3.1.0",
        "flask-cors==3.0.10",
        "gunicorn>20.0.0",
        "kubernetes==24.2.0",
        "marshmallow==2.19.5",
        "pyjwt==2.4.0",
        "requests-toolbelt==0.9.1",
        "sqlalchemy==1.3.22",
        "Cython==0.29.15",
        "Flask-SQLAlchemy==2.5.1",
        "Jinja2==3.0.3",
        "Werkzeug==2.2.3",
        "importlib-metadata==4.12.0",
        "pytest==7.0.0",
        "pytest-cov",
        "importanize==0.7.0",
        "PyYAML",
        "requests==2.26.0",
        "psycopg2-binary",
    ],

}

extras_require["complete"] = sorted(set(sum(extras_require.values(), [])))

extras_require["pmml"] = sorted(
    set([pack for pack in extras_require["complete"] if pack != "pyspark==2.4.6"])
)

extras_require["nb_api"] = sorted(
    set(extras_require["k8"] + ["GitPython==3.1.30","gitdb2==2.0.6"])
)

extras_require["pmml"].append("pypmml==0.9.5")

extras_require["rstudio"] = sorted(
    set([pack for pack in extras_require["complete"] if pack not in [
        "tensorflow==1.14.0",
        "tensorflow==2.4.1",
        "tensorflow-probability==0.11.1",
        "torch==1.2.0",
        "torchvision==0.4.0",
        "pandas==1.0.4"]])
)

not_require = [
    "keras==2.2.4",
    "pyspark==2.4.6",
    "tensorflow==2.4.1",
    "tensorflow==1.14.0",
    "tensorflow-probability==0.11.1",
    "torch==1.2.0",
    "torch==1.8.0",
    "torchvision==0.4.0",
    "torchvision==0.8.0",
    "spacy==2.2.2",
    "sklearn==0.0",
    "scikit-learn==0.21.3",
]


def make_lite(ele):
    return sorted(set(extras_require[ele]) - set(not_require))


extras_require["pmml_lite"] = make_lite("pmml")
extras_require["rstudio_lite"] = make_lite("rstudio")
extras_require["complete_lite"] = make_lite("complete")
extras_require["flavours_lite"] = make_lite("flavours")
extras_require["metrics_lite"] = make_lite("metrics")
extras_require["mosaicautoml_lite"] = make_lite("mosaicautoml")
extras_require["r-jupyter_lite"] = make_lite("r-jupyter")


setup(
    name="mosaic-common-utils",
    version="1.0.3",
    description="Utils library for Mosaic",
    url="https://git.lti-aiq.in/mosaic-ai-logistics/mosaic-common-utils",
    author="Rushikesh Raut",
    author_email="rushikesh.raut@lntinfotech.com",
    classifiers=["Programming Language :: Python :: 3.6"],
    packages=find_packages(),
    include_package_data=True,
    install_requires=[],
    extras_require=extras_require,
    )



# -*- coding: utf-8 -*-
import json
import numpy as np
import pandas as pd
import requests
import yaml
from sklearn import metrics
from sklearn.metrics import (
    accuracy_score,
    average_precision_score,
    brier_score_loss,
    confusion_matrix,
    f1_score,
    fbeta_score,
    log_loss,
    matthews_corrcoef,
    precision_score,
    recall_score,
    roc_curve,
    zero_one_loss,
    explained_variance_score,
    mean_absolute_error,
    mean_squared_error,
    mean_squared_log_error,
    r2_score,
    roc_auc_score,
)

from .utils import (
    get_headers,
    get_model_type_handler,
    try_or,
    get_array_type_handler,
    get_pred_and_true_values,
    update_progress,
)
from .constants import MosaicAI, MLModelMetricsV1


def update_metrics(pipeline_id, metrics, tag):
    payload = {"metric_value": json.dumps(metrics)}
    url = MosaicAI.server + MLModelMetricsV1.u.format(pipeline_id=pipeline_id, tag=tag)
    response = requests.put(url, json=payload, headers=get_headers())
    response.raise_for_status()


def get_estimator_obj(model_obj):
    if hasattr(model_obj, "estimators_"):
        try:
            model_estimator = model_obj.estimators_[0][0]
        except:
            model_estimator = model_obj.estimators_[0]
        return model_estimator


def metrics_stats(
    pipeline_id,
    ml_model_id,
    version_id,
    y_true,
    y_pred,
    prob,
    model_type,
    model_summary,
    model_obj,
    labels,
    features,
    original_features,
    encoded_columns,
):
    type_handler = get_model_type_handler(model_type)
    array_handler = get_array_type_handler(y_true, y_pred)
    if array_handler == "df":
        y_true, y_pred = get_pred_and_true_values(y_true, y_pred)
    all_matrix = type_handler(
        pipeline_id,
        ml_model_id,
        version_id,
        y_true,
        y_pred,
        prob,
        model_summary,
        model_type,
        model_obj,
        labels,
        features,
        original_features,
        encoded_columns,
    )
    return all_matrix


node_sum = 0


def tree_generator(
    clf, features, labels, original_features, node_index=0, side=0, prev_index=0
):
    global node_sum
    class_names = labels
    if node_sum <= 5000000:
        node = {}
        if clf.tree_.children_left[node_index] == -1:  # indicates leaf
            count_labels = zip(clf.tree_.value[node_index, 0], labels)
            node["pred"] = ", ".join(
                ("{} of {}".format(int(count), label) for count, label in count_labels)
            )

            node["side"] = "left" if side == "l" else "right"
            feature = features[clf.tree_.feature[prev_index]]
            threshold = clf.tree_.threshold[prev_index]

            if node_index == 0:
                node["name"] = "Root >"
            elif ("_-_" in feature) and (feature not in original_features):

                node["name"] = (
                    "{} = {}".format(feature.split("_-_")[0], feature.split("_-_")[1])
                    if side == "r"
                    else "{} != {}".format(
                        feature.split("_-_")[0], feature.split("_-_")[1]
                    )
                )
                node["type"] = "categorical"
            else:
                node["name"] = (
                    "{} > {}".format(feature, round(threshold, 2))
                    if side == "r"
                    else "{} <= {}".format(feature, round(threshold, 2))
                )
                node["type"] = "numerical"

            left_index = clf.tree_.children_left[node_index]
            right_index = clf.tree_.children_right[node_index]

            if clf.tree_.n_classes[0] != 1 and clf.tree_.n_outputs == 1:
                node["value"] = str(
                    class_names[np.argmax(clf.tree_.value[node_index, 0])]
                )
            else:
                node["value"] = sum(clf.tree_.value[node_index, 0])
        else:

            count_labels = zip(clf.tree_.value[node_index, 0], labels)
            node["pred"] = ", ".join(
                ("{} of {}".format(int(count), label) for count, label in count_labels)
            )

            node["side"] = "left" if side == "l" else "right"
            feature = features[clf.tree_.feature[prev_index]]
            threshold = clf.tree_.threshold[prev_index]

            if node_index == 0:
                node["name"] = "Root >"
            elif ("_-_" in feature) and (feature not in original_features):

                node["name"] = (
                    "{} = {}".format(feature.split("_-_")[0], feature.split("_-_")[1])
                    if side == "r"
                    else "{} != {}".format(
                        feature.split("_-_")[0], feature.split("_-_")[1]
                    )
                )
                node["type"] = "categorical"
            else:
                node["name"] = (
                    "{} > {}".format(feature, round(threshold, 2))
                    if side == "r"
                    else "{} <= {}".format(feature, round(threshold, 2))
                )
                node["type"] = "numerical"

            left_index = clf.tree_.children_left[node_index]
            right_index = clf.tree_.children_right[node_index]

            if clf.tree_.n_classes[0] != 1 and clf.tree_.n_outputs == 1:
                node["value"] = str(
                    class_names[np.argmax(clf.tree_.value[node_index, 0])]
                )
            else:
                node["value"] = sum(clf.tree_.value[node_index, 0])
            node_length = len(str(node))
            node_sum += node_length
            node["children"] = [
                tree_generator(
                    clf,
                    features,
                    labels,
                    original_features,
                    right_index,
                    "r",
                    node_index,
                ),
                tree_generator(
                    clf,
                    features,
                    labels,
                    original_features,
                    left_index,
                    "l",
                    node_index,
                ),
            ]
        return node


class MetricsBase:
    def __init__(
        self,
        pipeline_id,
        ml_model_id,
        version_id,
        y_true,
        y_pred,
        prob,
        model_info,
        model_type,
        model_obj,
        labels,
        features,
        original_features,
        encoded_columns,
    ):
        self.pipeline_id = pipeline_id
        self.ml_model_id = ml_model_id
        self.version_id = version_id
        self.y_true = y_true
        self.y_pred = y_pred
        self.prob = prob
        self.model_info = model_info
        self.model_type = model_type
        self.model_obj = model_obj
        self.labels = labels
        self.features = features
        self.original_features = original_features
        self.encoded_columns = encoded_columns
        self.progress_counter = 1
        self.progress_base_value = 12 if model_type == "classification" else 10

    def db_call(self, summary, tag):
        """
        Method for database operation

        Returns:
            json
        """
        if self.pipeline_id:
            payload = {
                "pipeline_id": self.pipeline_id,
                "metric_value": summary,
                "tag": tag,
            }
        else:
            payload = {
                "ml_model_id": self.ml_model_id,
                "version_id": self.version_id,
                "metric_value": summary,
                "tag": tag,
            }
        url = MosaicAI.server + MLModelMetricsV1.lc.format(version_id=self.version_id)
        response = requests.post(url, json=payload, headers=get_headers())
        response.raise_for_status()
        update_progress(self.progress_counter / self.progress_base_value)
        self.progress_counter += 1
        return response.json()

    def model_summary(self):
        summary_dict = {
            k: self.model_info[k] for k in list(self.model_info.keys())[:11]
        }
        model_dict = {"tag": "model_summary", "model_metric_value": summary_dict}
        update_progress(self.progress_counter / self.progress_base_value)
        self.progress_counter += 1
        return str(json.dumps(model_dict))

    def decision_tree(self):
        try:
            d_tree = tree_generator(
                self.model_obj,
                self.features,
                [self.labels, np.unique(self.y_pred)][self.labels is None],
                self.original_features,
            )
        except:
            try:
                d_tree = tree_generator(
                    get_estimator_obj(self.model_obj),
                    self.features,
                    [self.labels, np.unique(self.y_pred)][self.labels is None],
                    self.original_features,
                )
            except:
                d_tree = None
        decision_tree = json.dumps(d_tree).replace(", null", "").replace("null", "")
        model_metric_value = json.loads(decision_tree) if decision_tree else None
        decision_tree_dict = {
            "tag": "decision_tree",
            "model_metric_value": model_metric_value,
        }
        update_progress(self.progress_counter / self.progress_base_value)
        self.progress_counter += 1
        return str(json.dumps(decision_tree_dict))

    def cal_feature_importance(self):
        """
        Method to calculate variable importance

        Returns:
            String
        """
        try:
            index = self.encoded_columns if self.encoded_columns else self.features

            feature_imp = pd.Series(self.model_obj.feature_importances_, index=index)
            imp_list = []
            for index, value in feature_imp.items():
                d = {"column_name": index, "importance": value}
                imp_list.append(d.copy())
        except Exception:
            imp_list = None
        feature_importance_dict = {
            "tag": "feature_importance",
            "model_metric_value": imp_list,
        }
        update_progress(self.progress_counter / self.progress_base_value)
        self.progress_counter += 1
        return str(json.dumps(feature_importance_dict))


class Classification(MetricsBase):
    def __init__(
        self,
        pipeline_id,
        ml_model_id,
        version_id,
        y_true,
        y_pred,
        prob,
        model_info,
        model_type,
        model_obj,
        labels,
        features,
        original_features,
        encoded_columns,
    ):
        super().__init__(
            pipeline_id,
            ml_model_id,
            version_id,
            y_true,
            y_pred,
            prob,
            model_info,
            model_type,
            model_obj,
            labels,
            features,
            original_features,
            encoded_columns,
        )
        self.all_metrics()

    def all_metrics(self):
        """
        Calculate all metrics value to plot graph

        Returns:
            List
        """
        all_summary = {
            self.model_summary(),
            self.confusion_metrics(),
            self.decision_tree(),
            self.cal_roc_auc(),
            self.detailed_matrix(),
            self.cal_feature_importance(),
        }
        for summary in all_summary:
            tag = json.loads(summary)
            self.db_call(json.dumps(tag["model_metric_value"]), tag["tag"])
            update_progress(self.progress_counter / self.progress_base_value)
            self.progress_counter += 1
        return all_summary

    def confusion_metrics(self):
        """
        Method to return confusion matrix string

        Returns:
            String
        """
        try:
            dlist = []
            conf_matrix = confusion_matrix(self.y_true, self.y_pred)
            conf_list = conf_matrix.tolist()
            for idx, item in enumerate(conf_list):
                for index, val in enumerate(item):
                    d = {
                        "column_1_counter": idx,
                        "column_2_counter": index,
                        "prediction": val,
                        "column_1": idx,
                        "column_2": index,
                    }
                    dlist.append(d.copy())
        except Exception as ex:
            dlist = None
        conf_dict = {"tag": "confusion_matrix", "model_metric_value": dlist}
        update_progress(self.progress_counter / self.progress_base_value)
        self.progress_counter += 1
        return str(json.dumps(conf_dict))

    def detailed_matrix(self):
        """
        Method to calculate detailed matrix parameters

        Returns:
            String
        """
        accuracy_score_value = try_or(lambda: accuracy_score(self.y_true, self.y_pred))
        balanced_accuracy_score_value = try_or(
            lambda: metrics.balanced_accuracy_score(self.y_true, self.y_pred)
        )
        matthews_corrcoef_value = try_or(
            lambda: matthews_corrcoef(self.y_true, self.y_pred)
        )
        f1_score_value = try_or(
            lambda: f1_score(
                self.y_true,
                self.y_pred,
                average="weighted",
                labels=np.unique(self.y_pred),
            )
        )
        fbeta_score_value = try_or(
            lambda: fbeta_score(
                self.y_true,
                self.y_pred,
                average="weighted",
                beta=0.5,
                labels=np.unique(self.y_pred),
            )
        )
        precision_score_value = try_or(
            lambda: precision_score(
                self.y_true,
                self.y_pred,
                average="weighted",
                labels=np.unique(self.y_pred),
            )
        )
        recall_score_value = try_or(
            lambda: recall_score(self.y_true, self.y_pred, average="weighted")
        )
        zero_one_loss_value = try_or(lambda: zero_one_loss(self.y_true, self.y_pred))
        average_precision_score_value = try_or(
            lambda: average_precision_score(self.y_true, self.y_pred, pos_label=0)
        )
        log_loss_value = try_or(lambda: log_loss(self.y_true, self.prob))
        beier_score_loss_value = try_or(
            lambda: brier_score_loss(self.y_true, self.y_pred)
        )

        value_dict = {
            "Accuracy Score": accuracy_score_value,
            "Average Precision Score": average_precision_score_value,
            "Balanced Accuracy Score": balanced_accuracy_score_value,
            "Brier Score Loss": beier_score_loss_value,
            "Matthews Correlation Coefficient": matthews_corrcoef_value,
            "F1 Score": f1_score_value,
            "F-beta Score": fbeta_score_value,
            "Log Loss": log_loss_value,
            "Precision Score": precision_score_value,
            "Recall Score": recall_score_value,
            "Zero One Loss": zero_one_loss_value,
        }
        detailed_dict = {"tag": "detailed_matrix", "model_metric_value": value_dict}
        update_progress(self.progress_counter / self.progress_base_value)
        self.progress_counter += 1
        return str(json.dumps(detailed_dict))

    def cal_roc_auc(self):
        """
        Method to calculate roc auc curve parameters

        Returns:
            String
        """
        try:
            roc_auc = roc_auc_score(self.y_true, self.prob, average="macro")
            fpr, tpr, thresholds = roc_curve(self.y_true, self.prob)
            fpr = fpr.tolist()
            tpr = tpr.tolist()
            roc_auc_value = {"fpr": fpr, "tpr": tpr, "data": roc_auc}
        except Exception as ex:
            roc_auc_value = None
        roc_auc_dict = {"tag": "roc_auc", "model_metric_value": roc_auc_value}
        update_progress(self.progress_counter / self.progress_base_value)
        self.progress_counter += 1
        return str(json.dumps(roc_auc_dict))


class Regression(MetricsBase):
    def __init__(
        self,
        pipeline_id,
        ml_model_id,
        version_id,
        y_true,
        y_pred,
        prob,
        model_info,
        model_type,
        model_obj,
        labels,
        features,
        original_features,
        encoded_columns,
    ):
        super().__init__(
            pipeline_id,
            ml_model_id,
            version_id,
            y_true,
            y_pred,
            prob,
            model_info,
            model_type,
            model_obj,
            labels,
            features,
            original_features,
            encoded_columns,
        )
        self.all_matrix()

    def all_matrix(self):
        """
        Calculate all metrics value to plot graph

        Returns:
            List
        """
        all_summary = (
            self.model_summary(),
            self.cal_scatter_plot(get_array_type_handler(self.y_true, self.y_pred)),
            self.detailed_matrix(),
            self.decision_tree(),
            self.cal_feature_importance(),
        )
        for summary in all_summary:
            tag = yaml.safe_load(summary)
            self.db_call(json.dumps(tag["model_metric_value"]), tag["tag"])
            update_progress(self.progress_counter / self.progress_base_value)
            self.progress_counter += 1
        return all_summary

    def cal_scatter_plot(self, array_type):
        """
        Method to calculate parameters required for scatter plot
        Args:

        Returns:
            String
        """
        if not array_type == "list":
            y_true = self.y_true.ravel().tolist()
            y_pred = self.y_pred.ravel().tolist()
            list1 = [[f, b] for f, b in zip(y_true, y_pred)]
        else:
            list1 = [[f, b] for f, b in zip(self.y_true, self.y_pred)]
        scatt_dict = {"tag": "scatter_plot", "model_metric_value": list1}
        update_progress(self.progress_counter / self.progress_base_value)
        self.progress_counter += 1
        return str(json.dumps(scatt_dict))

    def mean_absolute_percentage_error(self):
        """
        Method to calculate mean absolute percentage error
        Args:

        Returns:
            String
        """
        y_true, y_pred = np.array(self.y_true), np.array(self.y_pred)
        # handling inifinity value for MAPE
        a = np.abs((y_true - y_pred) / y_true)
        return np.mean(a[np.isinf(a) == False]) * 100

    def detailed_matrix(self):
        """
        Method to calculate detailed matrix parameters
        Args:

        Returns:
            String
        """
        explained_variance_score_value = try_or(
            lambda: explained_variance_score(self.y_true, self.y_pred)
        )
        mean_absolute_error_value = try_or(
            lambda: mean_absolute_error(self.y_true, self.y_pred)
        )
        mean_absolute_percentage_error_value = try_or(
            lambda: self.mean_absolute_percentage_error()
        )
        mean_squared_error_value = try_or(
            lambda: mean_squared_error(self.y_true, self.y_pred)
        )
        rmse_value = try_or(
            lambda: np.sqrt(mean_squared_error(self.y_true, self.y_pred))
        )
        r2_score_value = try_or(lambda: r2_score(self.y_true, self.y_pred))
        root_mean_squared_log_error = try_or(
            lambda: np.sqrt(mean_squared_log_error(self.y_true, self.y_pred))
        )
        value_dict = {
            "Explained Variance Score": explained_variance_score_value,
            "Mean Absolute Error": mean_absolute_error_value,
            "Mean Absolute Percentage Error": mean_absolute_percentage_error_value,
            "Mean Squared Error": mean_squared_error_value,
            "Root Mean Square Error": rmse_value,
            "Root Mean Squared Logarithmic Error": root_mean_squared_log_error,
            "R2 score": r2_score_value,
        }
        dict = {"tag": "detailed_matrix", "model_metric_value": value_dict}
        update_progress(self.progress_counter / self.progress_base_value)
        self.progress_counter += 1
        return str(json.dumps(dict))



# -*- coding: utf-8 -*-

"""Zero Count."""

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils import check_array


class ZeroCount(BaseEstimator, TransformerMixin):
    """Adds the count of zeros and count of non-zeros per sample as features."""

    def fit(self, X, y=None):
        """Dummy function to fit in with the sklearn API."""
        return self

    def transform(self, X, y=None):
        """Transform data by adding two virtual features.

        Parameters
        ----------
        X: numpy ndarray, {n_samples, n_components}
            New data, where n_samples is the number of samples and n_components
            is the number of components.
        y: None
            Unused

        Returns
        -------
        X_transformed: array-like, shape (n_samples, n_features)
            The transformed feature set
        """
        X = check_array(X)
        n_features = X.shape[1]

        X_transformed = np.copy(X)

        non_zero_vector = np.count_nonzero(X_transformed, axis=1)
        non_zero = np.reshape(non_zero_vector, (-1, 1))
        zero_col = np.reshape(n_features - non_zero_vector, (-1, 1))

        X_transformed = np.hstack((non_zero, X_transformed))
        X_transformed = np.hstack((zero_col, X_transformed))

        return X_transformed



# -*- coding: utf-8 -*-
import os
import tarfile
from zipfile import ZipFile
from uuid import uuid4
import zipfile

def make_tar(files, output_file):
    tar = tarfile.open(output_file, "w:gz")
    for file in files:
        tar.add(file, arcname=os.path.basename(file))
    tar.close()


def extract_tar(tar_file, output_dir):
    tar = tarfile.open(tar_file)
    tar.extractall(path=output_dir)
    tar.close()


def extract_zip(zip_file, output_dir):
    with ZipFile(zip_file, "r") as f:
        f.extractall(path=output_dir)


def pickle_dumps(obj, file_path):
    import cloudpickle
    out = open(file_path, "wb")
    cloudpickle.dump(obj, out)
    out.close()


def pickle_loads(file_path):
    import cloudpickle
    from os import path
    if path.exists(file_path):
        pickled_file = open(file_path, "rb")
        obj = cloudpickle.load(pickled_file)
        pickled_file.close()
        return obj



def create_model_tar(base_dir, tar_name=None, *files):
    if tar_name is None:
        tar_path = os.path.join(base_dir, "ml_model.tar.gz")
    else:
        tar_path = os.path.join(base_dir, f"{tar_name}.tar.gz")
    make_tar(files, tar_path)
    return tar_path


def uuid_generator():
    _uuid = uuid4()
    return str(_uuid)


def check_tar_zip(file_path, temp_dir, file_name, file_type):
    if file_type=='application/gzip':
        new_filename = file_name.split(".")[0]
        extract_tar(file_path, temp_dir+f"/{new_filename}/")
        return "tar", temp_dir+f"/{new_filename}/"
    if file_type in ['application/zip', 'application/x-zip', 'application/octet-stream', 'application/x-zip-compressed']:
        new_filename = file_name.split(".")[0]
        extract_zip(file_path, temp_dir+f"/{new_filename}/")
        return "zip", temp_dir+f"/{new_filename}/"
    return "file", file_name


def check_and_create_directory(path):
    """
    Checking if folder exist if folder is not exist then it will create the folders
    :param path
    :return:file_path
    """
    try:
        path_array = path.split("/")
        directory = "/"
        for split_path in path_array:
            directory = directory + split_path + "/"
            if not os.path.isdir(directory):
                os.mkdir(directory, 0o777)
        os.chmod(path, 0o777)
        return "Folder created successfully!"
    except Exception as ex:
        raise ex


def save_model_data(file_path, obj_file):
    """
    Save data at File path
    :param file_path - destination path
    :param obj_file - file path / file obj to be stored on destination path
    :return: file_name
    """
    try:
        if not obj_file:
            return False
        if type(obj_file) is str:
            # File path has been passed
            # Creating required directories & copy file at the required location
            check_and_create_directory(file_path)
            import shutil
            shutil.copy2(obj_file, file_path)
            file_name = obj_file.split(os.path.sep)[-1]
            return file_name
        else:
            # File object has been passed
            # Creating required directories & stroing file at the required location
            base_path = '/'.join(file_path.split('/')[:-1])
            check_and_create_directory(base_path)
            file = open(file_path, "wb")
            for _chunk in obj_file.stream(32 * 1024):
                file.write(_chunk)
            file.close()
            return file_path.split('/')[-1]
    except Exception as ex:
        raise ex




from .constants import *
import requests


def audit_logging(console_url, action_type, object_id, object_name, object_type, headers, message=None, object_json=None):
    """
    Calls Audit Logging API of console backend

    Args:
        object_json (json): json response received from API
        headers (dict): headers to be passed to API
        object_type (string): MODEL or NOTEBOOK ( type of object )
        object_name (string): name of model or notebook
        object_id (string): id of model or notebook
        action_type (string): CREATE,UPDATE,DELETE,DEPLOY
        console_url (string): url of console backend API
        message (string) : custom message for API
    Returns:
        dict
    """
    # assigning headers for api
    headers = {
        Headers.x_auth_userid: headers["X-Auth-Userid"],
        Headers.x_auth_email: headers["X-Auth-Email"],
        Headers.x_auth_username: headers["X-Auth-Username"],
        Headers.x_project_id: headers["X-Project-Id"],
    }
    # json payload for audit logging
    payload = {
        "actionType": action_type,
        "objectId": object_id,
        "objectName": object_name,
        "objectType": object_type,
        "objectJson": object_json,
        "projectId": headers["X-Project-Id"],
        "message": message
    }
    # creating the api url
    audit_log_url = f"{console_url}/secured/api/activity/v1"
    # hitting the api
    response = requests.post(audit_log_url, headers=headers, json=payload)
    # printing response to check errors
    print(response.text)



# -*- coding: utf-8 -*-
import joblib
import json
from .utils import NumpyArrayEncoder


def load_model(model_path):
    return joblib.load(model_path)


def dump_model(model, path):
    joblib.dump(model, path)


def get_model_structure(model_obj):
    from math import isnan
    ml_model_class = model_obj.__dict__
    ml_model_class["class"] = str(model_obj.__class__)[8:-2:]
    ml_model_class = {k: (0 if isinstance(v, (float, int, list, dict, tuple)) and isnan(v) else v) for k, v in
                      ml_model_class.items()}
    return json.dumps(ml_model_class, cls=NumpyArrayEncoder)
